"""Implementation of Rule ST04."""

from typing import cast

from sqlfluff.core.parser import BaseSegment, Indent, NewlineSegment, WhitespaceSegment
from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.functional import FunctionalContext, Segments, sp
from sqlfluff.utils.reflow.reindent import construct_single_indent


class Rule_ST04(BaseRule):
    """Nested ``CASE`` statement in ``ELSE`` clause could be flattened.

    **Anti-pattern**

    In this example, the outer ``CASE``'s ``ELSE`` is an unnecessary, nested ``CASE``.

    .. code-block:: sql

        SELECT
          CASE
            WHEN species = 'Cat' THEN 'Meow'
            ELSE
            CASE
               WHEN species = 'Dog' THEN 'Woof'
            END
          END as sound
        FROM mytable

    **Best practice**

    Move the body of the inner ``CASE`` to the end of the outer one.

    .. code-block:: sql

        SELECT
          CASE
            WHEN species = 'Cat' THEN 'Meow'
            WHEN species = 'Dog' THEN 'Woof'
          END AS sound
        FROM mytable

    """

    name = "structure.nested_case"
    aliases = ("L058",)
    groups = ("all", "structure")
    crawl_behaviour = SegmentSeekerCrawler({"case_expression"})
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> LintResult:
        """Nested CASE statement in ELSE clause could be flattened."""
        segment = FunctionalContext(context).segment
        assert segment.select(sp.is_type("case_expression"))
        case1_children = segment.children()
        case1_first_case = case1_children.first(sp.is_keyword("CASE")).get()
        assert case1_first_case
        case1_first_when = case1_children.first(
            sp.is_type("when_clause", "else_clause")
        ).get()
        case1_last_when = case1_children.last(sp.is_type("when_clause")).get()
        case1_else_clause = case1_children.select(sp.is_type("else_clause"))
        case1_else_expressions = case1_else_clause.children(sp.is_type("expression"))
        expression_children = case1_else_expressions.children()
        case2 = expression_children.select(sp.is_type("case_expression"))
        case2_children = case2.children()
        case2_first_case = case2_children.first(sp.is_keyword("CASE")).get()
        case2_first_when = case2_children.first(
            sp.is_type("when_clause", "else_clause")
        ).get()
        # The len() checks below are for safety, to ensure the CASE inside
        # the ELSE is not part of a larger expression. In that case, it's
        # not safe to simplify in this way -- we'd be deleting other code.
        if (
            not case1_last_when
            or len(case1_else_expressions) > 1
            or len(expression_children) > 1
            or not case2
        ):
            return LintResult()

        # Determine if we can combine the else case statement, the first and
        # second case expressions should be the same. If they aren't, that
        # case currently isn't handled.
        if [
            x.raw_upper
            for x in segment.children(sp.is_code())
            .select(start_seg=case1_first_case, stop_seg=case1_first_when)
            .raw_segments
        ] != [
            x.raw_upper
            for x in case2.children(sp.is_code())
            .select(start_seg=case2_first_case, stop_seg=case2_first_when)
            .raw_segments
        ]:
            return LintResult()

        # We can assert that this exists because of the previous check.
        assert case1_last_when
        # We can also assert that we'll also have an else clause because
        # otherwise the case2 check above would fail.
        case1_else_clause_seg = case1_else_clause.get()
        assert case1_else_clause_seg

        # Delete stuff between the last "WHEN" clause and the "ELSE" clause.
        case1_to_delete = case1_children.select(
            start_seg=case1_last_when, stop_seg=case1_else_clause_seg
        )
        # Restore any comments that were deleted
        after_last_comment_index = (
            case1_to_delete.find(case1_to_delete.last(sp.is_comment()).get()) + 1
        )
        case1_comments_to_restore = case1_to_delete.select(
            stop_seg=case1_to_delete.get(after_last_comment_index)
        )
        after_else_comment = case1_else_clause.children().select(
            select_if=sp.is_type("newline", "comment", "whitespace"),
            stop_seg=case1_else_expressions.get(),
        )

        # Delete the nested "CASE" expression.
        fixes = case1_to_delete.apply(LintFix.delete)

        tab_space_size: int = context.config.get("tab_space_size", ["indentation"])
        indent_unit: str = context.config.get("indent_unit", ["indentation"])

        # Determine the indentation to use when we move the nested "WHEN"
        # and "ELSE" clauses, based on the indentation of case1_last_when.
        # If no whitespace segments found, use default indent.
        when_indent_str = self._get_indentation(
            case1_children, case1_last_when, tab_space_size, indent_unit
        )
        # Again determine indentation, but matching the "CASE"/"END" level.
        end_indent_str = self._get_indentation(
            case1_children, case1_first_case, tab_space_size, indent_unit
        )

        # Move the nested "when" and "else" clauses after the last outer
        # "when".
        nested_clauses = case2.children(
            sp.is_type("when_clause", "else_clause", "newline", "comment", "whitespace")
        )

        # Rebuild the nested case statement.
        # Any comments after the last outer "WHEN" that were deleted
        segments = list(case1_comments_to_restore)
        # Any comments between the "ELSE" and nested "CASE"
        segments += self._rebuild_spacing(when_indent_str, after_else_comment)
        # The nested "WHEN", "ELSE" or "comments", with logical spacing
        segments += self._rebuild_spacing(when_indent_str, nested_clauses)
        fixes.append(LintFix.create_after(case1_last_when, segments, source=segments))

        # Delete the outer "else" clause.
        fixes.append(LintFix.delete(case1_else_clause_seg))
        # Add spacing for any comments that may exist after the nested `END`
        # but only on that same line.
        fixes += self._nested_end_trailing_comment(
            case1_children, case1_else_clause_seg, end_indent_str
        )
        return LintResult(case2[0], fixes=fixes)

    def _get_indentation(
        self,
        parent_segments: Segments,
        segment: BaseSegment,
        tab_space_size: int,
        indent_unit: str,
    ) -> str:
        """Calculate the indentation level for rebuilding nested struct.

        This is only a best attempt as the input may not be equally indented. The layout
        rules, if run, would resolve this.
        """
        leading_whitespace = (
            parent_segments.select(stop_seg=segment)
            .reversed()
            .first(sp.is_type("whitespace"))
        )
        seg_indent = parent_segments.select(stop_seg=segment).last(sp.is_type("indent"))
        indent_level = 1
        if (
            seg_indent
            and (segment_indent := seg_indent.get())
            and segment_indent.is_type("indent")
        ):
            # is_type("indent") matches both Indent and Dedent (Dedent subclasses
            # Indent); the cast is for typing only — `indent_val` exists on both
            # `Indent` and the RsSegment façade at runtime.
            indent_level = cast(Indent, segment_indent).indent_val + 1
        indent_str = (
            "".join(seg.raw for seg in leading_whitespace)
            if leading_whitespace
            and (whitespace_seg := leading_whitespace.get())
            and len(whitespace_seg.raw) > 1
            else construct_single_indent(indent_unit, tab_space_size) * indent_level
        )

        return indent_str

    def _nested_end_trailing_comment(
        self,
        case1_children: Segments,
        case1_else_clause_seg: BaseSegment,
        end_indent_str: str,
    ) -> list[LintFix]:
        """Prepend newline spacing to comments on the final nested `END` line."""
        trailing_end = case1_children.select(
            start_seg=case1_else_clause_seg,
            loop_while=sp.not_(sp.is_type("newline")),
        )
        fixes = trailing_end.select(
            sp.is_whitespace(), loop_while=sp.not_(sp.is_comment())
        ).apply(LintFix.delete)
        first_comment = trailing_end.first(sp.is_comment()).get()
        if first_comment:
            segments = [NewlineSegment(), WhitespaceSegment(end_indent_str)]
            fixes.append(LintFix.create_before(first_comment, segments, segments))
        return fixes

    def _rebuild_spacing(
        self, indent_str: str, nested_clauses: Segments
    ) -> list[BaseSegment]:
        buff = []
        # If the first segment is a comment, add a newline
        prior_newline = nested_clauses.first(sp.not_(sp.is_whitespace())).any(
            sp.is_comment()
        )
        prior_whitespace = ""
        for seg in nested_clauses:
            if seg.is_type("when_clause", "else_clause") or (
                prior_newline and seg.is_comment
            ):
                buff += [NewlineSegment(), WhitespaceSegment(indent_str), seg]
                prior_newline = False
                prior_whitespace = ""
            elif seg.is_type("newline"):
                prior_newline = True
                prior_whitespace = ""
            elif not prior_newline and seg.is_comment:
                buff += [WhitespaceSegment(prior_whitespace), seg]
                prior_newline = False
                prior_whitespace = ""
            elif seg.is_whitespace:
                # Don't reset newline
                prior_whitespace = seg.raw
        return buff



"""Implementation of Rule LT03."""

from collections.abc import Sequence

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.reflow import ReflowSequence


class Rule_LT03(BaseRule):
    """Operators should follow a standard for being before/after newlines.

    The configuration for whether operators should be ``trailing`` or
    ``leading`` is part of :ref:`layoutconfig`. The default configuration is:

    .. code-block:: cfg

        [sqlfluff:layout:type:binary_operator]
        line_position = leading

        [sqlfluff:layout:type:comparison_operator]
        line_position = leading

    **Anti-pattern**

    In this example, if ``line_position = leading`` (or unspecified, as is the
    default), then the operator ``+`` should not be at the end of the second line.

    .. code-block:: sql

        SELECT
            a +
            b
        FROM foo


    **Best practice**

    If ``line_position = leading`` (or unspecified, as this is the default),
    place the operator after the newline.

    .. code-block:: sql

        SELECT
            a
            + b
        FROM foo

    If ``line_position = trailing``, place the operator before the newline.

    .. code-block:: sql

        SELECT
            a +
            b
        FROM foo
    """

    name = "layout.operators"
    aliases = ("L007",)
    groups = ("all", "layout")
    crawl_behaviour = SegmentSeekerCrawler(
        {
            "binary_operator",
            "comparison_operator",
            "assignment_operator",
            "pipe_operator",
        }
    )
    is_fix_compatible = True

    def _seek_newline(
        self, segments: Sequence[BaseSegment], idx: int, dir: int
    ) -> bool:
        """Seek in a direction, looking for newlines.

        Args:
            segments: A sequence of segments to seek within.
            idx: The index of the "current" segment.
            dir: The direction to seek in (+1 for forward, -1 for backward)
        """
        assert dir in (1, -1)
        for segment in segments[idx + dir :: dir]:
            if segment.is_type("newline"):
                # It's definitely leading. No problems.
                self.logger.debug(
                    "Shortcut (dir = %s) OK. Found newline: %s", dir, segment
                )
                return True
            elif not segment.is_type("whitespace", "indent", "comment"):
                # We found something before it which suggests it's not leading.
                # We should run the full reflow routine to check.
                break
        return False

    def _check_trail_lead_shortcut(
        self, segment: BaseSegment, parent: BaseSegment, line_position: str
    ) -> bool:
        """Check to see whether we should pass the rule and shortcut.

        Args:
            segment: The target segment.
            parent: The parent segment (must contain `segment`).
            line_position: The `line_position` config for the segment.
        """
        base_position = line_position.split(":")[0]
        if base_position not in ("leading", "trailing"):
            return False
        attached = "attached" in line_position
        strict = line_position.endswith("strict")

        # `leading`/`trailing` are mirror images, so resolve the sides relative
        # to the desired position and share the logic.
        idx = parent.segments.index(segment)
        has_newline_before = self._seek_newline(parent.segments, idx, dir=-1)
        has_newline_after = self._seek_newline(parent.segments, idx, dir=1)
        if base_position == "leading":
            in_position, opposite = has_newline_before, has_newline_after
        else:
            in_position, opposite = has_newline_after, has_newline_before

        if attached:
            # `attached` also forbids floating alone on its own line.
            if opposite:
                return False
            return in_position or not strict
        # OK unless mispositioned; `strict` also rejects mid-line.
        return in_position or (not strict and not opposite)

    def _eval(self, context: RuleContext) -> list[LintResult]:
        """Operators should follow a standard for being before/after newlines.

        For the fixing routines we delegate to the reflow utils. However
        for performance reasons we have some initial shortcuts to quickly
        identify situations which are _ok_ to avoid the overhead of the
        full reflow path.
        """
        # NOTE: These shortcuts assume that any newlines will be direct
        # siblings of the operator in question. This isn't _always_ the case
        # but is true often enough to have meaningful upside from early
        # detection. `pipe_operator` has no shortcut and falls through to reflow.
        for operator_type in (
            "comparison_operator",
            "binary_operator",
            "assignment_operator",
        ):
            if context.segment.is_type(operator_type):
                positioning = context.config.get(
                    "line_position", ["layout", "type", operator_type]
                )
                if self._check_trail_lead_shortcut(
                    context.segment, context.parent_stack[-1], positioning
                ):
                    return [LintResult()]
                break

        return (
            ReflowSequence.from_around_target(
                context.segment,
                root_segment=context.parent_stack[0],
                config=context.config,
            )
            .rebreak()
            .get_results()
        )



"""Implementation of Rule PG02."""

from typing import Optional

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


def _has_keyword(segments, keyword: str) -> bool:
    """Check if a keyword exists among direct child segments."""
    return any(seg.is_type("keyword") and seg.raw_upper == keyword for seg in segments)


class Rule_PG02(BaseRule):
    """Create PostgreSQL foreign keys as ``NOT VALID`` before validating.

    Adding a foreign key constraint normally validates existing rows as part of
    the ``ALTER TABLE`` statement. On large tables, this can hold locks for
    longer than necessary. Creating the constraint as ``NOT VALID`` defers that
    scan so it can be run later with ``VALIDATE CONSTRAINT``.

    This rule only applies to the ``postgres`` dialect and is disabled by
    default. Enable it with the ``force_enable = True`` flag.

    **Anti-pattern**

    A foreign key constraint that validates existing rows during creation.

    .. code-block:: sql

        ALTER TABLE foo ADD CONSTRAINT fk_bar
            FOREIGN KEY (bar_id) REFERENCES bar (id);

    **Best practice**

    Create the foreign key as ``NOT VALID``, then validate it in a separate
    statement.

    .. code-block:: sql

        ALTER TABLE foo ADD CONSTRAINT fk_bar
            FOREIGN KEY (bar_id) REFERENCES bar (id) NOT VALID;

        ALTER TABLE foo VALIDATE CONSTRAINT fk_bar;

    """

    name = "postgres.not_valid_foreign_key"
    aliases = ()
    groups = ("all", "postgres")
    config_keywords = ["force_enable"]
    crawl_behaviour = SegmentSeekerCrawler({"alter_table_statement"})

    _target_dialects = ("postgres",)

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        self.force_enable: bool

        if not self.force_enable:
            return None

        if context.dialect.name not in self._target_dialects:
            return None

        assert context.segment.is_type("alter_table_statement")

        action = context.segment.get_child("alter_table_action_segment")
        if not action:  # pragma: no cover
            return None
        constraint = action.get_child("table_constraint")
        if not constraint:
            return None
        if not _has_keyword(constraint.segments, "FOREIGN"):
            return None
        if _has_keyword(constraint.segments, "VALID"):
            return None
        return LintResult(
            anchor=context.segment,
            description=(
                "ADD CONSTRAINT ... FOREIGN KEY should use NOT VALID, then "
                "VALIDATE CONSTRAINT separately, to avoid locking the table "
                "while validating existing rows."
            ),
        )



"""Errors - these are closely linked to what used to be called violations.

NOTE: The BaseException class, which ValueError inherits from, defines
a custom __reduce__() method for picking and unpickling exceptions.
For the SQLBaseError, and it's dependent classes, we define properties
of these exceptions which don't work well with that method, which is
why we redefine __reduce__() on each of these classes. Given the
circumstances in which they are called, they don't show up on coverage
tracking.

https://stackoverflow.com/questions/49715881/how-to-pickle-inherited-exceptions
"""

from typing import TYPE_CHECKING, Any, Optional, Union, cast

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.parser import BaseSegment, PositionMarker
    from sqlfluff.core.rules import BaseRule, LintFix

    try:
        from sqlfluffrs import RsParseError, RsSQLLexerError
    except ImportError:
        ...


CheckTuple = tuple[str, int, int]
SerializedObject = dict[str, Union[str, int, bool, list["SerializedObject"]]]


def _extract_position(segment: Optional["BaseSegment"]) -> dict[str, int]:
    """If a segment is present and is a literal, return it's source length."""
    if segment:
        position = segment.pos_marker
        assert position
        if position.is_literal():
            return position.to_source_dict()
    # An empty location is an indicator of not being able to accurately
    # represent the location.
    return {}  # pragma: no cover


class SQLBaseError(ValueError):
    """Base Error Class for all violations."""

    _code: Optional[str] = None
    _name: Optional[str] = None
    _identifier = "base"
    _warning = False  # The default value for `warning`

    def __init__(
        self,
        description: Optional[str] = None,
        pos: Optional["PositionMarker"] = None,
        line_no: int = 0,
        line_pos: int = 0,
        ignore: bool = False,
        fatal: bool = False,
        warning: Optional[bool] = None,
    ) -> None:
        self.fatal = fatal
        self.ignore = ignore
        self.warning: bool = warning if warning is not None else self._warning
        self.description = description
        if pos:
            self.line_no, self.line_pos = pos.source_position()
        else:
            self.line_no = line_no
            self.line_pos = line_pos
        super().__init__(self.desc())

    def __eq__(self, other: Any) -> bool:
        """Errors compare equal if they are the same type and same content."""
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __reduce__(
        self,
    ) -> tuple[type["SQLBaseError"], tuple[Any, ...]]:
        """Prepare the SQLBaseError for pickling."""
        return type(self), (
            self.description,
            None,
            self.line_no,
            self.line_pos,
            self.ignore,
            self.fatal,
            self.warning,
        )

    @property
    def fixable(self) -> bool:
        """Should this error be considered fixable?"""
        return False

    def rule_code(self) -> str:
        """Fetch the code of the rule which cause this error."""
        return self._code or "????"

    def rule_name(self) -> str:
        """Fetch the name of the rule which cause this error."""
        return self._name or "????"

    def desc(self) -> str:
        """Fetch a description of this violation."""
        if self.description:
            return self.description

        return self.__class__.__name__  # pragma: no cover

    def to_dict(self) -> SerializedObject:
        """Return a dict of properties.

        This is useful in the API for outputting violations.
        """
        return {
            "start_line_no": self.line_no,
            "start_line_pos": self.line_pos,
            "code": self.rule_code(),
            "description": self.desc(),
            "name": getattr(self, "rule").name if hasattr(self, "rule") else "",
            "warning": self.warning,
        }

    def check_tuple(self) -> CheckTuple:
        """Get a tuple representing this error. Mostly for testing."""
        return (
            self.rule_code(),
            self.line_no,
            self.line_pos,
        )

    def source_signature(self) -> tuple[Any, ...]:
        """Return hashable source signature for deduplication."""
        return (self.check_tuple(), self.desc())

    def ignore_if_in(self, ignore_iterable: list[str]) -> None:
        """Ignore this violation if it matches the iterable."""
        if self._identifier in ignore_iterable:
            self.ignore = True

    def warning_if_in(self, warning_iterable: list[str]) -> None:
        """Warning only for this violation if it matches the iterable.

        Designed for rule codes so works with L001, LL0X but also TMP or PRS
        for templating and parsing errors.

        Args:
            warning_iterable (list[str]): A list of strings representing the warning
                codes to check.

        Returns:
            None
        """
        if self.rule_code() in warning_iterable or self.rule_name() in warning_iterable:
            self.warning = True


class SQLTemplaterError(SQLBaseError):
    """An error which occurred during templating.

    Args:
        pos (:obj:`PosMarker`, optional): The position which the error
            occurred at.

    """

    _code = "TMP"
    _identifier = "templating"


class SQLFluffSkipFile(RuntimeError):
    """An error returned from a templater to skip a file."""

    pass


class SQLLexError(SQLBaseError):
    """An error which occurred during lexing.

    Args:
        pos (:obj:`PosMarker`, optional): The position which the error
            occurred at.

    """

    _code = "LXR"
    _identifier = "lexing"

    @classmethod
    def from_rs_error(cls, rs_error: "RsSQLLexerError") -> "SQLLexError":
        """Create a SQLLexError from a RsSQLLexerError."""
        return cls(
            description=rs_error.desc,
            line_no=rs_error.line_no,
            line_pos=rs_error.line_pos,
            ignore=rs_error.ignore,
            fatal=rs_error.fatal,
            warning=rs_error.warning,
        )


class SQLParseError(SQLBaseError):
    """An error which occurred during parsing.

    Args:
        segment (:obj:`BaseSegment`, optional): The segment which is relevant
            for the failure in parsing. This is likely to be a subclass of
            `BaseSegment` rather than the parent class itself. This is mostly
            used for logging and for referencing position.

    """

    _code = "PRS"
    _identifier = "parsing"

    def __init__(
        self,
        description: Optional[str] = None,
        segment: Optional["BaseSegment"] = None,
        line_no: int = 0,
        line_pos: int = 0,
        ignore: bool = False,
        fatal: bool = False,
        warning: Optional[bool] = None,
    ) -> None:
        # Store the segment on creation - we might need it later
        self.segment = segment
        super().__init__(
            description=description,
            pos=segment.pos_marker if segment else None,
            line_no=line_no,
            line_pos=line_pos,
            ignore=ignore,
            fatal=fatal,
            warning=warning,
        )

    @classmethod
    def from_rs_parse_error(
        cls,
        rs_error: "RsParseError",
        segments: tuple["BaseSegment", ...],
    ) -> "SQLParseError":
        """Convert a Rust RsParseError to SQLParseError.

        Args:
            rs_error: The RsParseError exception from Rust parser
            segments: The segment array being parsed

        Returns:
            SQLParseError with position information extracted from rs_error
        """
        error_msg = str(rs_error)
        error_segment: Optional["BaseSegment"] = None
        error_pos = getattr(rs_error, "pos", 0)
        error_segment = segments[int(error_pos or 0)]

        return cls(
            description=error_msg,
            segment=error_segment,
        )

    def __reduce__(
        self,
    ) -> tuple[type["SQLParseError"], tuple[Any, ...]]:
        """Prepare the SQLParseError for pickling."""
        return type(self), (
            self.description,
            self.segment,
            self.line_no,
            self.line_pos,
            self.ignore,
            self.fatal,
            self.warning,
        )

    def to_dict(self) -> SerializedObject:
        """Return a dict of properties.

        This is useful in the API for outputting violations.

        For parsing errors we additionally add the length of the unparsable segment.
        """
        _base_dict = super().to_dict()
        _base_dict.update(
            **_extract_position(self.segment),
        )
        return _base_dict


class SQLLintError(SQLBaseError):
    """An error which occurred during linting.

    In particular we reference the rule here to do extended logging based on
    the rule in question which caused the fail.

    Args:
        segment (:obj:`BaseSegment`, optional): The segment which is relevant
            for the failure in parsing. This is likely to be a subclass of
            `BaseSegment` rather than the parent class itself. This is mostly
            used for logging and for referencing position.

    """

    _identifier = "linting"

    def __init__(
        self,
        description: str,
        segment: "BaseSegment",
        rule: "BaseRule",
        fixes: Optional[list["LintFix"]] = None,
        ignore: bool = False,
        fatal: bool = False,
        warning: Optional[bool] = None,
    ) -> None:
        self.segment = segment
        self.rule = rule
        self.fixes = fixes or []
        super().__init__(
            description=description,
            pos=segment.pos_marker if segment else None,
            ignore=ignore,
            fatal=fatal,
            warning=warning,
        )

    def __reduce__(
        self,
    ) -> tuple[type["SQLLintError"], tuple[Any, ...]]:
        """Prepare the SQLLintError for pickling."""
        return type(self), (
            self.description,
            self.segment,
            self.rule,
            self.fixes,
            self.ignore,
            self.fatal,
            self.warning,
        )

    def to_dict(self) -> SerializedObject:
        """Return a dict of properties.

        This is useful in the API for outputting violations.

        For linting errors we additionally add details of any fixes.
        """
        _base_dict = super().to_dict()
        _base_dict.update(
            fixes=[fix.to_dict() for fix in self.fixes],
            **_extract_position(self.segment),
        )
        # Edge case: If the base error doesn't have an end position
        # but we only have one fix and it _does_. Then use use that in the
        # overall fix.
        _fixes = cast(list[SerializedObject], _base_dict.get("fixes", []))
        if "end_line_pos" not in _base_dict and len(_fixes) == 1:
            _fix = _fixes[0]
            # If the mandatory keys match...
            if (
                _fix["start_line_no"] == _base_dict["start_line_no"]
                and _fix["start_line_pos"] == _base_dict["start_line_pos"]
            ):
                # ...then hoist all the optional ones from the fix.
                for key in [
                    "start_file_pos",
                    "end_line_no",
                    "end_line_pos",
                    "end_file_pos",
                ]:
                    _base_dict[key] = _fix[key]

        return _base_dict

    @property
    def fixable(self) -> bool:
        """Should this error be considered fixable?"""
        if self.fixes:
            return True
        return False

    def rule_code(self) -> str:
        """Fetch the code of the rule which cause this error."""
        return self.rule.code

    def rule_name(self) -> str:
        """Fetch the name of the rule which cause this error."""
        return self.rule.name

    def source_signature(self) -> tuple[Any, ...]:
        """Return hashable source signature for deduplication.

        For linting errors we need to dedupe on more than just location and
        description, we also need to check the edits potentially made, both
        in the templated file but also in the source.
        """
        fix_raws = tuple(
            tuple(e.raw for e in f.edit) if f.edit else None for f in self.fixes
        )
        _source_fixes: list[tuple[str, int, int]] = []
        for fix in self.fixes:
            if not fix.edit:
                continue
            for edit in fix.edit:
                for source_edit in edit.source_fixes:
                    # NOTE: It's important that we don't dedupe on the
                    # templated slice for the source fix, because that will
                    # be different for different locations in any loop.
                    _source_fixes.append(
                        (
                            source_edit.edit,
                            source_edit.source_slice.start,
                            source_edit.source_slice.stop,
                        )
                    )
        return (self.check_tuple(), self.description, fix_raws, tuple(_source_fixes))

    def __repr__(self) -> str:
        return "<SQLLintError: rule {} pos:{!r}, #fixes: {}, description: {}>".format(
            self.rule_code(),
            (self.line_no, self.line_pos),
            len(self.fixes),
            self.description,
        )


class SQLUnusedNoQaWarning(SQLBaseError):
    """A warning about an unused noqa directive."""

    _code = "NOQA"
    _identifier = "noqa"
    _warning = True


class SQLFluffUserError(ValueError):
    """An error which should be fed back to the user."""



"""Implementation of Rule OR01."""

from typing import Optional

from sqlfluff.core.rules import BaseRule, LintFix, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler


class Rule_OR01(BaseRule):
    """Remove empty batches.

    **Anti-pattern**

    Empty batches (containing only / statements) should be removed.

    .. code-block:: sql
       :force:

        SELECT 1 FROM DUAL;

        /

        /

    **Best practice**

    Remove empty batches.

    .. code-block:: sql
       :force:

        SELECT 1 FROM DUAL;

        /
    """

    name = "oracle.empty_batch"
    aliases = ()
    groups = ("all", "oracle")
    crawl_behaviour = SegmentSeekerCrawler({"batch"})
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[list[LintResult]]:
        """Remove empty batches."""
        # Rule only applies to Oracle syntax.
        if context.dialect.name != "oracle":
            return None  # pragma: no cover

        if not context.segment.is_type("batch"):  # pragma: no cover
            return None

        # Get all non-whitespace, non-meta segments
        content_segments = [
            s
            for s in context.segment.segments
            if not s.is_type("whitespace", "newline", "indent", "dedent")
            and not s.is_meta
        ]

        # Check if batch contains only slash_buffer_executor statement
        has_real_content = False
        has_slash = False
        for content_seg in content_segments:
            if content_seg.is_type("slash_buffer_executor"):
                has_slash = True
            elif not content_seg.is_meta:
                has_real_content = True
                break

        # If batch has no real content (only / and whitespace), remove it
        if not has_real_content and has_slash:
            fixes = [LintFix.delete(context.segment)]

            # Also delete the trailing newline after the batch if present
            segments = context.parent_stack[-1].segments
            idx = segments.index(context.segment)
            if idx + 1 < len(segments):
                next_seg = segments[idx + 1]
                if next_seg.is_type("newline"):
                    fixes.append(LintFix.delete(next_seg))

            return [
                LintResult(
                    anchor=context.segment,
                    description="Empty batch with only / statement should be removed.",
                    fixes=fixes,
                )
            ]

        return None



"""File Helpers for the parser module."""

import codecs
import os.path
from collections.abc import Iterator
from pathlib import Path
from typing import Optional

import chardet


def get_encoding(fname: str, config_encoding: str = "autodetect") -> str:
    """Get the encoding of the file (autodetect)."""
    if config_encoding != "autodetect":
        return config_encoding

    with open(fname, "rb") as f:
        data = f.read()

    if data.startswith((codecs.BOM_UTF32_BE, codecs.BOM_UTF32_LE)):
        return "utf-32"
    if data.startswith((codecs.BOM_UTF16_BE, codecs.BOM_UTF16_LE)):
        return "utf-16"
    if data.startswith(codecs.BOM_UTF8):
        return "utf-8-sig"

    if all(char < 128 for char in data):
        return "ascii"

    detected_encoding = chardet.detect(data).get("encoding")
    if not detected_encoding:
        return "utf-8"
    return detected_encoding


def iter_intermediate_paths(inner_path: Path, outer_path: Path) -> Iterator[Path]:
    """Iterate paths between two given paths.

    If the `inner_path` is a subdirectory of the `outer_path` then all steps
    in between the two are yielded as Path objects, from outer to inner including
    the two at each end. If not, then the just the `outer_path` and `inner_path`
    are returned (in that order).

    NOTE: The current behaviour is not currently precisely as described above.
    Instead, we look for the lowest *common path* between the inner and outer
    paths. This is a superset of the originally intended logic, but is convenient
    until we have a good solution for the dbt templater project root path.

    * If there is not common path, the outer path and inner path are yielded *only*.
    * If there is a common path, then that common path is yielded first, and then
      paths leading up to the inner path are yielded. Unless the inner path is a
      subdirectory of the outer path, the *outer path is not yielded*.

    In both scenarios, the inner path is always the last path to be yielded.
    """
    inner_path = inner_path.absolute()
    outer_path = outer_path.absolute()

    # If we've been passed a file and not a directory,
    # then go straight to the directory.
    # NOTE: We only check this for the inner path.
    if not inner_path.is_dir():
        inner_path = inner_path.parent

    common_path: Optional[Path]
    try:
        common_path = Path(os.path.commonpath([inner_path, outer_path])).absolute()
    except ValueError:
        # Getting a value error means that we're likely on a windows system
        # and have been provided a `inner_path` and `outer_path` which are
        # in different drives. In this situation, there's no shared path,
        # so just yield the given path.
        common_path = None

    # NOTE: I think the following logic here isn't correct. It is too expansive
    # in the search locations for config files. Correcting that without access
    # to the root project location for a dbt project and therefore allowing a
    # a more accurate search is not feasible. In future that path should somehow
    # be made available here.

    if not common_path:
        yield outer_path.resolve()
    else:
        # we have a sub path! We can load nested paths
        path_to_visit = common_path
        while path_to_visit != inner_path:
            yield path_to_visit.resolve()
            next_path_to_visit = (
                path_to_visit / inner_path.relative_to(path_to_visit).parts[0]
            )
            if next_path_to_visit == path_to_visit:  # pragma: no cover
                # we're not making progress...
                # [prevent infinite loop]
                break
            path_to_visit = next_path_to_visit

    yield inner_path.resolve()



"""The sqlite dialect.

https://www.sqlite.org/
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    AnyNumberOf,
    Anything,
    BaseSegment,
    Bracketed,
    CodeSegment,
    CommentSegment,
    Dedent,
    Delimited,
    IdentifierSegment,
    Indent,
    LiteralSegment,
    Matchable,
    NewlineSegment,
    Nothing,
    OneOf,
    OptionallyBracketed,
    ParseMode,
    Ref,
    RegexLexer,
    Sequence,
    StringLexer,
    StringParser,
    SymbolSegment,
    TypedParser,
    WhitespaceSegment,
)
from sqlfluff.core.parser.grammar.lookbehind import is_distinct_from_lookbehind
from sqlfluff.dialects import dialect_ansi as ansi
from sqlfluff.dialects.dialect_sqlite_keywords import (
    RESERVED_KEYWORDS,
    UNRESERVED_KEYWORDS,
)

ansi_dialect = load_raw_dialect("ansi")

sqlite_dialect = ansi_dialect.copy_as(
    "sqlite",
    formatted_name="SQLite",
    docstring="""**Default Casing**: Not specified in the docs,
but through testing it appears that SQLite *stores* column names
in whatever case they were defined, but is always *case-insensitive*
when resolving those names.

**Quotes**: String Literals: ``''`` (or  ``""`` if not otherwise resolved
to an identifier), Identifiers: ``""``, ``[]`` or |back_quotes|. See the
`SQLite Keywords Docs`_ for more details.

The dialect for `SQLite <https://www.sqlite.org/>`_.

.. _`SQLite Keywords Docs`: https://sqlite.org/lang_keywords.html
""",
)

sqlite_dialect.sets("reserved_keywords").clear()
sqlite_dialect.sets("reserved_keywords").update(RESERVED_KEYWORDS)
sqlite_dialect.sets("unreserved_keywords").clear()
sqlite_dialect.sets("unreserved_keywords").update(UNRESERVED_KEYWORDS)

sqlite_dialect.patch_lexer_matchers(
    [
        # SQLite allows block comments to be terminated by end of input
        RegexLexer(
            "block_comment",
            r"\/\*([^\*]|\*(?!\/))*(\*\/|\Z)",
            CommentSegment,
            subdivider=RegexLexer(
                "newline",
                r"\r\n|\n",
                NewlineSegment,
            ),
            trim_post_subdivide=RegexLexer(
                "whitespace",
                r"[^\S\r\n]+",
                WhitespaceSegment,
            ),
        ),
        RegexLexer(
            "single_quote",
            r"'([^']|'')*'",
            CodeSegment,
            segment_kwargs={
                "quoted_value": (r"'((?:[^']|'')*)'", 1),
                "escape_replacements": [(r"''", "'")],
            },
        ),
        RegexLexer(
            "double_quote",
            r'"([^"]|"")*"',
            CodeSegment,
            segment_kwargs={
                "quoted_value": (r'"((?:[^"]|"")*)"', 1),
                "escape_replacements": [(r'""', '"')],
            },
        ),
        RegexLexer(
            "back_quote",
            r"`([^`]|``)*`",
            CodeSegment,
            segment_kwargs={
                "quoted_value": (r"`((?:[^`]|``)*)`", 1),
                "escape_replacements": [(r"``", "`")],
            },
        ),
    ]
)

sqlite_dialect.insert_lexer_matchers(
    [
        RegexLexer(
            "blob_literal",
            r"[xX]'([\da-fA-F][\da-fA-F])*'",
            LiteralSegment,
            segment_kwargs={"type": "blob_literal"},
        ),
        RegexLexer(
            "hexadecimal_literal",
            r"0x[\da-fA-F]+",
            LiteralSegment,
            segment_kwargs={"type": "numeric_literal"},
        ),
        RegexLexer(
            "at_sign_literal",
            r"@[a-zA-Z0-9_]+",
            LiteralSegment,
            segment_kwargs={"type": "at_sign_literal"},
        ),
        RegexLexer(
            "colon_literal",
            r":[a-zA-Z0-9_]+",
            LiteralSegment,
            segment_kwargs={"type": "colon_literal"},
        ),
        RegexLexer(
            "question_literal",
            r"\?[0-9]+",
            LiteralSegment,
            segment_kwargs={"type": "question_literal"},
        ),
        RegexLexer(
            "dollar_literal",
            r"\$[a-zA-Z0-9_]+",
            LiteralSegment,
            segment_kwargs={"type": "dollar_literal"},
        ),
    ],
    before="question",
)

sqlite_dialect.insert_lexer_matchers(
    [
        StringLexer("inline_path_operator", "->>", CodeSegment),
        StringLexer("column_path_operator", "->", CodeSegment),
    ],
    before="greater_than",
)

sqlite_dialect.add(
    BackQuotedIdentifierSegment=TypedParser(
        "back_quote",
        IdentifierSegment,
        type="quoted_identifier",
        # match ANSI's naked identifier casefold, sqlite is case-insensitive.
        casefold=str.upper,
    ),
    ColumnPathOperatorSegment=StringParser(
        "->", SymbolSegment, type="column_path_operator"
    ),
    InlinePathOperatorSegment=StringParser(
        "->>", SymbolSegment, type="column_path_operator"
    ),
    QuestionMarkSegment=StringParser("?", SymbolSegment, type="question_mark"),
    AtSignLiteralSegment=TypedParser(
        "at_sign_literal",
        LiteralSegment,
        type="at_sign_literal",
    ),
    ColonLiteralSegment=TypedParser(
        "colon_literal",
        LiteralSegment,
        type="colon_literal",
    ),
    QuestionLiteralSegment=TypedParser(
        "question_literal",
        LiteralSegment,
        type="question_literal",
    ),
    DollarLiteralSegment=TypedParser(
        "dollar_literal",
        LiteralSegment,
        type="dollar_literal",
    ),
    BlobLiteralSegment=TypedParser(
        "blob_literal",
        LiteralSegment,
        type="blob_literal",
    ),
)

sqlite_dialect.replace(
    PrimaryKeyGrammar=Sequence(
        "PRIMARY",
        "KEY",
        OneOf("ASC", "DESC", optional=True),
        Ref("ConflictClauseSegment", optional=True),
        Sequence("AUTOINCREMENT", optional=True),
    ),
    NumericLiteralSegment=OneOf(
        TypedParser("numeric_literal", LiteralSegment, type="numeric_literal"),
        Ref("ParameterizedSegment"),
    ),
    LiteralGrammar=ansi_dialect.get_grammar("LiteralGrammar").copy(
        insert=[Ref("ParameterizedSegment"), Ref("BlobLiteralSegment")]
    ),
    TemporaryTransientGrammar=Ref("TemporaryGrammar"),
    DateTimeLiteralGrammar=Sequence(
        OneOf("DATE", "DATETIME"),
        TypedParser("single_quote", LiteralSegment, type="date_constructor_literal"),
    ),
    BaseExpressionElementGrammar=OneOf(
        Ref("LiteralGrammar"),
        Ref("BareFunctionSegment"),
        Ref("FunctionSegment"),
        Ref("ColumnReferenceSegment"),
        Ref("ExpressionSegment"),
        Sequence(
            Ref("DatatypeSegment"),
            Ref("LiteralGrammar"),
        ),
        terminators=[
            Ref("CommaSegment"),
            Ref.keyword("AS"),
        ],
    ),
    AlterTableOptionsGrammar=OneOf(
        Sequence("RENAME", "TO", Ref("SingleIdentifierGrammar")),
        Sequence(
            "RENAME",
            Sequence("COLUMN", optional=True),
            Ref("ColumnReferenceSegment"),
            "TO",
            Ref("SingleIdentifierGrammar"),
        ),
        Sequence(
            "ADD", Sequence("COLUMN", optional=True), Ref("ColumnDefinitionSegment")
        ),
        Sequence(
            "DROP", Sequence("COLUMN", optional=True), Ref("ColumnReferenceSegment")
        ),
    ),
    AutoIncrementGrammar=Nothing(),
    CommentClauseSegment=Nothing(),
    IntervalExpressionSegment=Nothing(),
    TimeZoneGrammar=Nothing(),
    FetchClauseSegment=Nothing(),
    TrimParametersGrammar=Nothing(),
    LikeGrammar=Sequence("LIKE"),
    OverlapsClauseSegment=Nothing(),
    MLTableExpressionSegment=Nothing(),
    MergeIntoLiteralGrammar=Nothing(),
    SamplingExpressionSegment=Nothing(),
    BinaryOperatorGrammar=ansi_dialect.get_grammar("BinaryOperatorGrammar").copy(
        insert=[
            Ref("ColumnPathOperatorSegment"),
            Ref("InlinePathOperatorSegment"),
        ]
    ),
    OrderByClauseTerminators=OneOf(
        "LIMIT",
        # For window functions
        "WINDOW",
        Ref("FrameClauseUnitGrammar"),
    ),
    WhereClauseTerminatorGrammar=OneOf(
        "LIMIT",
        Sequence("GROUP", "BY"),
        Sequence("ORDER", "BY"),
        "WINDOW",
    ),
    FromClauseTerminatorGrammar=OneOf(
        "WHERE",
        "LIMIT",
        Sequence("GROUP", "BY"),
        Sequence("ORDER", "BY"),
        "WINDOW",
        Ref("SetOperatorSegment"),
        Ref("WithNoSchemaBindingClauseSegment"),
        Ref("WithDataClauseSegment"),
    ),
    GroupByClauseTerminatorGrammar=OneOf(
        Sequence("ORDER", "BY"),
        "LIMIT",
        "HAVING",
        "WINDOW",
    ),
    PostFunctionGrammar=Sequence(
        Ref("FilterClauseGrammar", optional=True),
        Ref("OverClauseSegment", optional=True),
    ),
    IgnoreRespectNullsGrammar=Nothing(),
    SelectClauseTerminatorGrammar=OneOf(
        Ref(
            "FromKeywordSegment",
            exclude=is_distinct_from_lookbehind,
        ),
        "WHERE",
        Sequence("ORDER", "BY"),
        "LIMIT",
        Ref("SetOperatorSegment"),
    ),
    FunctionContentsGrammar=AnyNumberOf(
        Ref("ExpressionSegment"),
        # A Cast-like function
        Sequence(Ref("ExpressionSegment"), "AS", Ref("DatatypeSegment")),
        # Trim function
        Sequence(
            Ref("TrimParametersGrammar"),
            Ref("ExpressionSegment", optional=True, exclude=Ref.keyword("FROM")),
            "FROM",
            Ref("ExpressionSegment"),
        ),
        # An extract-like or substring-like function
        Sequence(
            OneOf(Ref("DatetimeUnitSegment"), Ref("ExpressionSegment")),
            "FROM",
            Ref("ExpressionSegment"),
        ),
        Sequence(
            # Allow an optional distinct keyword here.
            Ref.keyword("DISTINCT", optional=True),
            OneOf(
                # Most functions will be using the delimited route
                # but for COUNT(*) or similar we allow the star segment
                # here.
                Ref("StarSegment"),
                Delimited(Ref("FunctionContentsExpressionGrammar")),
            ),
        ),
        Ref(
            "OrderByClauseSegment"
        ),  # used by string_agg (postgres), group_concat (exasol),listagg (snowflake)..
        # like a function call: POSITION ( 'QL' IN 'SQL')
        Sequence(
            OneOf(
                Ref("QuotedLiteralSegment"),
                Ref("SingleIdentifierGrammar"),
                Ref("ColumnReferenceSegment"),
            ),
            "IN",
            OneOf(
                Ref("QuotedLiteralSegment"),
                Ref("SingleIdentifierGrammar"),
                Ref("ColumnReferenceSegment"),
            ),
        ),
        Ref("IndexColumnDefinitionSegment"),
        # Raise Function contents
        OneOf(
            "IGNORE",
            Sequence(
                OneOf(
                    "ABORT",
                    "FAIL",
                    "ROLLBACK",
                ),
                Ref("CommaSegment"),
                Ref("QuotedLiteralSegment"),
            ),
        ),
    ),
    # NOTE: This block was copy/pasted from dialect_ansi.py with these changes:
    #  - "PRIOR" keyword removed from Expression_A_Unary_Operator_Grammar
    #  - QuantifiedComparisonOperatorGrammar removed (SQLite doesn't support
    #    ANY/ALL/SOME)
    Expression_A_Unary_Operator_Grammar=OneOf(
        Ref(
            "SignedSegmentGrammar",
            exclude=Sequence(Ref("QualifiedNumericLiteralSegment")),
        ),
        Ref("TildeSegment"),
        Ref("NotOperatorGrammar"),
    ),
    Expression_A_Grammar=Sequence(
        Ref("Tail_Recurse_Expression_A_Grammar"),
        AnyNumberOf(
            OneOf(
                Ref("LikeExpressionGrammar"),
                Sequence(
                    Ref("BinaryOperatorGrammar"),
                    Ref("Tail_Recurse_Expression_A_Grammar"),
                ),
                Ref("InOperatorGrammar"),
                # QuantifiedComparisonOperatorGrammar removed - not supported in SQLite
                Sequence(
                    "IS",
                    Ref.keyword("NOT", optional=True),
                    Ref("IsClauseGrammar"),
                ),
                Ref("IsNullGrammar"),
                Ref("NotNullGrammar"),
                Ref("CollateGrammar"),
                Sequence(
                    Ref.keyword("NOT", optional=True),
                    "BETWEEN",
                    Ref("Expression_B_Grammar"),
                    "AND",
                    Ref("Tail_Recurse_Expression_A_Grammar"),
                ),
                Sequence(
                    Ref("PatternMatchingGrammar"),
                    Ref("Expression_A_Grammar"),
                ),
            )
        ),
    ),
    IsDistinctFromGrammar=Sequence(
        "IS",
        Ref.keyword("NOT", optional=True),
        Sequence("DISTINCT", "FROM", optional=True),
    ),
    NanLiteralSegment=Nothing(),
    PatternMatchingGrammar=Sequence(
        Ref.keyword("NOT", optional=True),
        OneOf("GLOB", "REGEXP", "MATCH"),
    ),
    SingleIdentifierGrammar=OneOf(
        Ref("NakedIdentifierSegment"),
        Ref("SingleQuotedIdentifierSegment"),
        Ref("QuotedIdentifierSegment"),
        Ref("BackQuotedIdentifierSegment"),
        terminators=[Ref("DotSegment")],
    ),
    # match ANSI's naked identifier casefold, sqlite is case-insensitive.
    QuotedIdentifierSegment=TypedParser(
        "double_quote", IdentifierSegment, type="quoted_identifier", casefold=str.upper
    ),
    SingleQuotedIdentifierSegment=TypedParser(
        "single_quote", IdentifierSegment, type="quoted_identifier", casefold=str.upper
    ),
    ColumnConstraintDefaultGrammar=Ref("ExpressionSegment"),
    FrameClauseUnitGrammar=OneOf("ROWS", "RANGE", "GROUPS"),
)


class FrameClauseSegment(BaseSegment):
    """A frame clause for window functions.

    https://www.sqlite.org/syntax/frame-spec.html
    """

    type = "frame_clause"

    match_grammar: Matchable = Sequence(
        Ref("FrameClauseUnitGrammar"),
        OneOf(
            Sequence("UNBOUNDED", "PRECEDING"),
            Sequence("CURRENT", "ROW"),
            Sequence(Ref("ExpressionSegment"), "PRECEDING"),
            Sequence(
                "BETWEEN",
                OneOf(
                    Sequence("UNBOUNDED", "PRECEDING"),
                    Sequence("CURRENT", "ROW"),
                    Sequence(Ref("ExpressionSegment"), "FOLLOWING"),
                    Sequence(Ref("ExpressionSegment"), "PRECEDING"),
                ),
                "AND",
                OneOf(
                    Sequence("UNBOUNDED", "FOLLOWING"),
                    Sequence("CURRENT", "ROW"),
                    Sequence(Ref("ExpressionSegment"), "FOLLOWING"),
                    Sequence(Ref("ExpressionSegment"), "PRECEDING"),
                ),
            ),
        ),
        Sequence(
            "EXCLUDE",
            OneOf(
                Sequence("NO", "OTHERS"), Sequence("CURRENT", "ROW"), "TIES", "GROUP"
            ),
            optional=True,
        ),
    )


class ParameterizedSegment(BaseSegment):
    """Sqlite allows named and argument based parameters to prevent SQL Injection.

    https://www.sqlite.org/c3ref/bind_blob.html

    """

    type = "parameterized_expression"
    match_grammar = OneOf(
        Ref("AtSignLiteralSegment"),
        Ref("QuestionMarkSegment"),
        Ref("ColonLiteralSegment"),
        Ref("QuestionLiteralSegment"),
        Ref("DollarLiteralSegment"),
    )


class SetOperatorSegment(BaseSegment):
    """A set operator such as Union, Minus, Except or Intersect."""

    type = "set_operator"
    match_grammar: Matchable = OneOf(
        Sequence("UNION", OneOf("DISTINCT", "ALL", optional=True)),
        Sequence(
            OneOf(
                "INTERSECT",
                "EXCEPT",
            ),
            Ref.keyword("ALL", optional=True),
        ),
        exclude=Sequence("EXCEPT", Bracketed(Anything())),
    )


class ColumnReferenceSegment(ansi.ColumnReferenceSegment):
    """A reference to column, field or alias.

    Also allows `column->path` and `column->>path` for JSON values.
    https://www.sqlite.org/json1.html#jptr
    """

    match_grammar = ansi.ColumnReferenceSegment.match_grammar.copy(
        insert=[
            Sequence(
                OneOf(
                    ansi.ColumnReferenceSegment.match_grammar.copy(),
                    Ref("FunctionSegment"),
                    Ref("BareFunctionSegment"),
                    Ref("LiteralGrammar"),
                ),
            ),
        ]
    )


class TableReferenceSegment(ansi.TableReferenceSegment):
    """A reference to a table.

    Also allows `table->path` and `table->>path` for JSON values.
    https://www.sqlite.org/json1.html#jptr
    """

    match_grammar = ansi.TableReferenceSegment.match_grammar.copy(
        insert=[
            Sequence(
                ansi.TableReferenceSegment.match_grammar.copy(),
                OneOf(
                    Ref("ColumnPathOperatorSegment"),
                    Ref("InlinePathOperatorSegment"),
                ),
                OneOf(
                    Ref("LiteralGrammar"),
                ),
            ),
        ]
    )


class DatatypeSegment(ansi.DatatypeSegment):
    """A data type segment.

    Supports timestamp with(out) time zone. Doesn't currently support intervals.
    """

    type = "data_type"
    match_grammar: Matchable = OneOf(
        Sequence(
            "DOUBLE",
            "PRECISION",
        ),
        Sequence("UNSIGNED", "BIG", "INT"),
        Sequence(
            OneOf(
                Sequence(
                    OneOf("VARYING", "NATIVE"),
                    OneOf("CHARACTER"),
                ),
                Sequence(
                    OneOf("CHARACTER"),
                    OneOf("VARYING", "NATIVE"),
                ),
                Ref("DatatypeIdentifierSegment"),
            ),
            Ref("BracketedArguments", optional=True),
            OneOf("UNSIGNED", optional=True),
        ),
    )


class TableEndClauseSegment(BaseSegment):
    """Support Table Options at end of tables.

    https://www.sqlite.org/syntax/table-options.html
    """

    type = "table_end_clause_segment"
    match_grammar: Matchable = Delimited(Sequence("WITHOUT", "ROWID"), "STRICT")


class ValuesClauseSegment(ansi.ValuesClauseSegment):
    """A `VALUES` clause like in `INSERT`."""

    type = "values_clause"
    match_grammar: Matchable = Sequence(
        "VALUES",
        Delimited(
            Sequence(
                Bracketed(
                    Delimited(
                        "DEFAULT",
                        Ref("ExpressionSegment"),
                    ),
                    parse_mode=ParseMode.GREEDY,
                ),
            ),
        ),
    )


class IndexColumnDefinitionSegment(BaseSegment):
    """A column definition for CREATE INDEX.

    Overridden from ANSI to allow expressions
    https://www.sqlite.org/expridx.html.
    """

    type = "index_column_definition"
    match_grammar: Matchable = Sequence(
        OneOf(
            Ref("SingleIdentifierGrammar"),  # Column name
            Ref("ExpressionSegment"),  # Expression for simple functions
        ),
        OneOf("ASC", "DESC", optional=True),
    )


class ReturningClauseSegment(BaseSegment):
    """A returning clause.

    Per docs https://www.sqlite.org/lang_returning.html
    """

    type = "returning_clause"

    match_grammar = Sequence(
        "RETURNING",
        Indent,
        Delimited(
            Ref("WildcardExpressionSegment"),
            Sequence(
                Ref("ExpressionSegment"),
                Ref("AliasExpressionSegment", optional=True),
            ),
        ),
        Dedent,
    )


class ConflictTargetSegment(BaseSegment):
    """An upsert conflict target.

    https://www.sqlite.org/lang_upsert.html
    """

    type = "conflict_target"
    match_grammar = Sequence(
        Delimited(Ref("IndexColumnDefinitionSegment")),
        Sequence("WHERE", Ref("ExpressionSegment"), optional=True),
    )


class UpsertClauseSegment(BaseSegment):
    """An upsert clause.

    https://www.sqlite.org/lang_upsert.html
    """

    type = "upsert_clause"
    match_grammar = Sequence(
        "ON",
        "CONFLICT",
        Ref("ConflictTargetSegment", optional=True),
        "DO",
        OneOf(
            "NOTHING",
            Sequence(
                "UPDATE",
                "SET",
                Indent,
                Delimited(
                    Sequence(
                        OneOf(
                            Ref("SingleIdentifierGrammar"),
                            Ref("BracketedColumnReferenceListGrammar"),
                        ),
                        Ref("EqualsSegment"),
                        Ref("ExpressionSegment"),
                    ),
                ),
                Dedent,
                Sequence(
                    "WHERE",
                    Ref("ExpressionSegment"),
                    optional=True,
                ),
            ),
        ),
    )


class InsertStatementSegment(BaseSegment):
    """An`INSERT` statement.

    https://www.sqlite.org/lang_insert.html
    """

    type = "insert_statement"
    match_grammar = Sequence(
        OneOf(
            Sequence(
                "INSERT",
                Sequence(
                    "OR",
                    OneOf(
                        "ABORT",
                        "FAIL",
                        "IGNORE",
                        "REPLACE",
                        "ROLLBACK",
                    ),
                    optional=True,
                ),
            ),
            # REPLACE is just an alias for INSERT OR REPLACE
            "REPLACE",
        ),
        "INTO",
        Ref("TableReferenceSegment"),
        Ref("BracketedColumnReferenceListGrammar", optional=True),
        OneOf(
            Sequence(
                Ref("ValuesClauseSegment"),
                Ref("UpsertClauseSegment", optional=True),
            ),
            Sequence(
                OptionallyBracketed(Ref("SelectableGrammar")),
                Ref("UpsertClauseSegment", optional=True),
            ),
            Ref("DefaultValuesGrammar"),
        ),
        Ref("ReturningClauseSegment", optional=True),
    )


class ConflictClauseSegment(BaseSegment):
    """A conflict clause.

    https://www.sqlite.org/lang_conflict.html
    """

    type = "conflict_clause"
    match_grammar = Sequence(
        "ON",
        "CONFLICT",
        OneOf(
            "ROLLBACK",
            "ABORT",
            "FAIL",
            "IGNORE",
            "REPLACE",
        ),
    )


class ColumnConstraintSegment(ansi.ColumnConstraintSegment):
    """A column option; each CREATE TABLE column can have 0 or more.

    Overriding ColumnConstraintSegment to allow for additional segment parsing
    and to support on conflict clauses.
    """

    match_grammar: Matchable = Sequence(
        Sequence(
            "CONSTRAINT",
            Ref("ObjectReferenceSegment"),  # Constraint name
            optional=True,
        ),
        OneOf(
            Sequence(
                Ref.keyword("NOT", optional=True),
                "NULL",
                Ref("ConflictClauseSegment", optional=True),
            ),  # NOT NULL or NULL
            Sequence("CHECK", Bracketed(Ref("ExpressionSegment"))),
            Sequence(  # DEFAULT <value>
                "DEFAULT",
                Ref("ColumnConstraintDefaultGrammar"),
            ),
            Ref("PrimaryKeyGrammar"),
            Sequence(
                Ref("UniqueKeyGrammar"), Ref("ConflictClauseSegment", optional=True)
            ),  # UNIQUE
            Ref("AutoIncrementGrammar"),
            Ref("ReferenceDefinitionGrammar"),  # REFERENCES reftable [ ( refcolumn) ]x
            Ref("CommentClauseSegment"),
            Sequence(
                "COLLATE", Ref("CollationReferenceSegment")
            ),  # https://www.sqlite.org/datatype3.html#collation
            Sequence(
                Sequence("GENERATED", "ALWAYS", optional=True),
                "AS",
                Bracketed(Ref("ExpressionSegment")),
                OneOf("STORED", "VIRTUAL", optional=True),
            ),  # https://www.sqlite.org/gencol.html
        ),
        OneOf("DEFERRABLE", Sequence("NOT", "DEFERRABLE"), optional=True),
        OneOf(
            Sequence("INITIALLY", "DEFERRED"),
            Sequence("INITIALLY", "IMMEDIATE"),
            optional=True,
        ),
    )


class TableConstraintSegment(ansi.TableConstraintSegment):
    """Overriding TableConstraintSegment to allow for additional segment parsing."""

    match_grammar: Matchable = Sequence(
        Sequence(  # [ CONSTRAINT <Constraint name> ]
            "CONSTRAINT", Ref("ObjectReferenceSegment"), optional=True
        ),
        OneOf(
            # CHECK ( <expr> )
            Sequence("CHECK", Bracketed(Ref("ExpressionSegment"))),
            Sequence(  # UNIQUE ( column_name [, ... ] )
                "UNIQUE",
                Ref("BracketedColumnReferenceListGrammar"),
                # Later add support for index_parameters?
                Ref("ConflictClauseSegment", optional=True),
            ),
            Sequence(  # PRIMARY KEY ( column_name [, ... ] ) index_parameters
                Ref("PrimaryKeyGrammar"),
                # Columns making up PRIMARY KEY constraint
                Ref("BracketedColumnReferenceListGrammar"),
                # Later add support for index_parameters?
                Ref("ConflictClauseSegment", optional=True),
            ),
            Sequence(  # FOREIGN KEY ( column_name [, ... ] )
                # REFERENCES reftable [ ( refcolumn [, ... ] ) ]
                Ref("ForeignKeyGrammar"),
                # Local columns making up FOREIGN KEY constraint
                Ref("BracketedColumnReferenceListGrammar"),
                Ref(
                    "ReferenceDefinitionGrammar"
                ),  # REFERENCES reftable [ ( refcolumn) ]
            ),
        ),
        OneOf("DEFERRABLE", Sequence("NOT", "DEFERRABLE"), optional=True),
        OneOf(
            Sequence("INITIALLY", "DEFERRED"),
            Sequence("INITIALLY", "IMMEDIATE"),
            optional=True,
        ),
    )


class TransactionStatementSegment(ansi.TransactionStatementSegment):
    """A `COMMIT`, `ROLLBACK` or `TRANSACTION` statement.

    As per https://www.sqlite.org/lang_transaction.html
    """

    type = "transaction_statement"
    match_grammar: Matchable = Sequence(
        OneOf("BEGIN", "COMMIT", "ROLLBACK", "END"),
        OneOf("TRANSACTION", optional=True),
        Sequence("TO", "SAVEPOINT", Ref("ObjectReferenceSegment"), optional=True),
    )


class PragmaReferenceSegment(ansi.ObjectReferenceSegment):
    """A Pragma object."""

    type = "pragma_reference"


class PragmaStatementSegment(BaseSegment):
    """A Pragma Statement.

    As per https://www.sqlite.org/pragma.html
    """

    type = "pragma_statement"

    _pragma_value = OneOf(
        Ref("LiteralGrammar"),
        Ref("BooleanLiteralGrammar"),
        "YES",
        "NO",
        "ON",
        "OFF",
        "NONE",
        "FULL",
        "INCREMENTAL",
        "DELETE",
        "TRUNCATE",
        "PERSIST",
        "MEMORY",
        "WAL",
        "NORMAL",
        "EXCLUSIVE",
        "FAST",
        "EXTRA",
        "DEFAULT",
        "FILE",
        "PASSIVE",
        "RESTART",
        "RESET",
    )

    match_grammar = Sequence(
        "PRAGMA",
        Ref("PragmaReferenceSegment"),
        OneOf(
            Bracketed(_pragma_value),
            Bracketed(Ref("ObjectReferenceSegment")),
            optional=True,
        ),
        Sequence(
            Ref("EqualsSegment"), OptionallyBracketed(_pragma_value), optional=True
        ),
    )


class CreateTriggerStatementSegment(ansi.CreateTriggerStatementSegment):
    """Create Trigger Statement.

    https://www.sqlite.org/lang_createtrigger.html
    """

    type = "create_trigger"

    match_grammar: Matchable = Sequence(
        "CREATE",
        Ref("TemporaryGrammar", optional=True),
        "TRIGGER",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TriggerReferenceSegment"),
        OneOf("BEFORE", "AFTER", Sequence("INSTEAD", "OF"), optional=True),
        OneOf(
            "DELETE",
            "INSERT",
            Sequence(
                "UPDATE",
                Sequence(
                    "OF",
                    Delimited(
                        Ref("ColumnReferenceSegment"),
                    ),
                    optional=True,
                ),
            ),
        ),
        "ON",
        Ref("TableReferenceSegment"),
        Sequence("FOR", "EACH", "ROW", optional=True),
        Sequence(
            "WHEN",
            Indent,
            OptionallyBracketed(Ref("ExpressionSegment")),
            Dedent,
            optional=True,
        ),
        "BEGIN",
        Indent,
        Delimited(
            Ref("UpdateStatementSegment"),
            Ref("InsertStatementSegment"),
            Ref("DeleteStatementSegment"),
            Ref("SelectableGrammar"),
            delimiter=AnyNumberOf(Ref("DelimiterGrammar"), min_times=1),
            allow_gaps=True,
            allow_trailing=True,
        ),
        Dedent,
        "END",
    )


class CreateViewStatementSegment(BaseSegment):
    """A `CREATE VIEW` statement."""

    type = "create_view_statement"
    # https://www.sqlite.org/lang_createview.html
    match_grammar: Matchable = Sequence(
        "CREATE",
        Ref("TemporaryGrammar", optional=True),
        "VIEW",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        # Optional list of column names
        Ref("BracketedColumnReferenceListGrammar", optional=True),
        "AS",
        OptionallyBracketed(Ref("SelectableGrammar")),
    )


class UnorderedSelectStatementSegment(BaseSegment):
    """A `SELECT` statement without any ORDER clauses or later.

    Replaces (without overriding) ANSI to remove Eager Matcher
    """

    type = "select_statement"

    match_grammar = Sequence(
        Ref("SelectClauseSegment"),
        Ref("FromClauseSegment", optional=True),
        Ref("WhereClauseSegment", optional=True),
        Ref("GroupByClauseSegment", optional=True),
        Ref("HavingClauseSegment", optional=True),
        Ref("OverlapsClauseSegment", optional=True),
        Ref("NamedWindowSegment", optional=True),
    )


class DeleteStatementSegment(ansi.DeleteStatementSegment):
    """A `DELETE` statement.

    DELETE FROM <table name> [ WHERE <search condition> ]
    """

    type = "delete_statement"
    # match grammar. This one makes sense in the context of knowing that it's
    # definitely a statement, we just don't know what type yet.
    match_grammar: Matchable = Sequence(
        "DELETE",
        Ref("FromClauseSegment"),
        Ref("WhereClauseSegment", optional=True),
        Ref("ReturningClauseSegment", optional=True),
    )


class UpdateStatementSegment(ansi.UpdateStatementSegment):
    """An `Update` statement.

    UPDATE <table name> SET <set clause list> [ WHERE <search condition> ]
    """

    type = "update_statement"
    match_grammar: Matchable = Sequence(
        "UPDATE",
        Sequence(
            "OR",
            OneOf(
                "ABORT",
                "FAIL",
                "IGNORE",
                "REPLACE",
                "ROLLBACK",
            ),
            optional=True,
        ),
        Indent,
        Ref("TableReferenceSegment"),
        Ref("AliasExpressionSegment", optional=True),
        Dedent,
        Ref("SetClauseListSegment"),
        Ref("FromClauseSegment", optional=True),
        Ref("WhereClauseSegment", optional=True),
        Ref("ReturningClauseSegment", optional=True),
    )


class SetClauseSegment(ansi.SetClauseSegment):
    """A set clause."""

    match_grammar = Sequence(
        OneOf(
            Ref("SingleIdentifierGrammar"),
            Ref("BracketedColumnReferenceListGrammar"),
        ),
        Ref("EqualsSegment"),
        Ref("ExpressionSegment"),
    )


class SelectStatementSegment(BaseSegment):
    """A `SELECT` statement.

    Replaces (without overriding) ANSI to remove Eager Matcher
    """

    type = "select_statement"
    # Remove the Limit and Window statements from ANSI
    match_grammar = UnorderedSelectStatementSegment.match_grammar.copy(
        insert=[
            Ref("OrderByClauseSegment", optional=True),
            Ref("FetchClauseSegment", optional=True),
            Ref("LimitClauseSegment", optional=True),
            Ref("NamedWindowSegment", optional=True),
        ]
    )


class GroupingSetsClauseSegment(ansi.GroupingSetsClauseSegment):
    """`GROUPING SETS` clause within the `GROUP BY` clause.

    This is `Nothing` for SQLite.
    """

    match_grammar = Nothing()


class CreateIndexStatementSegment(ansi.CreateIndexStatementSegment):
    """A `CREATE INDEX` statement.

    As per https://www.sqlite.org/lang_createindex.html
    """

    type = "create_index_statement"
    match_grammar: Matchable = Sequence(
        "CREATE",
        Ref.keyword("UNIQUE", optional=True),
        "INDEX",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("IndexReferenceSegment"),
        "ON",
        Ref("TableReferenceSegment"),
        Sequence(
            Bracketed(
                Delimited(
                    Ref("IndexColumnDefinitionSegment"),
                ),
            )
        ),
        Ref("WhereClauseSegment", optional=True),
    )


class CreateVirtualTableStatementSegment(BaseSegment):
    """A `CREATE VIRTUAL TABLE` statement.

    As per https://www.sqlite.org/lang_createvtab.html
    """

    type = "create_virtual_table_statement"
    match_grammar: Matchable = Sequence(
        "CREATE",
        "VIRTUAL",
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        "USING",
        Ref("SingleIdentifierGrammar"),
        Bracketed(
            Delimited(
                OneOf(
                    Sequence(
                        Ref("SingleIdentifierGrammar"),
                        Ref("EqualsSegment"),
                        OneOf(
                            Ref("QuotedLiteralSegment"),
                            Ref("NumericLiteralSegment"),
                            Ref("ObjectReferenceSegment"),
                        ),
                    ),
                    Ref("QuotedLiteralSegment"),
                    Ref("NumericLiteralSegment"),
                    Ref("SingleIdentifierGrammar"),
                ),
            ),
            optional=True,
        ),
    )


class StatementSegment(ansi.StatementSegment):
    """Overriding StatementSegment to allow for additional segment parsing."""

    match_grammar = OneOf(
        Ref("AlterTableStatementSegment"),
        Ref("CreateIndexStatementSegment"),
        Ref("CreateTableStatementSegment"),
        Ref("CreateVirtualTableStatementSegment"),
        Ref("CreateTriggerStatementSegment"),
        Ref("CreateViewStatementSegment"),
        Ref("DeleteStatementSegment"),
        Ref("DropIndexStatementSegment"),
        Ref("DropTableStatementSegment"),
        Ref("DropTriggerStatementSegment"),
        Ref("DropViewStatementSegment"),
        Ref("ExplainStatementSegment"),
        Ref("InsertStatementSegment"),
        Ref("PragmaStatementSegment"),
        Ref("SelectableGrammar"),
        Ref("TransactionStatementSegment"),
        Ref("UpdateStatementSegment"),
        Bracketed(Ref("StatementSegment")),
    )


class CTEDefinitionSegment(ansi.CTEDefinitionSegment):
    """A CTE Definition from a WITH statement.

    https://sqlite.org/lang_with.html

    Data-Modifying Statements (INSERT, UPDATE, DELETE) in WITH are
    matched by ansi.SelectableGrammar > NonWithSelectableGrammar.
    """

    match_grammar = Sequence(
        Ref("SingleIdentifierGrammar"),
        Ref("CTEColumnList", optional=True),
        "AS",
        Sequence(Ref.keyword("NOT", optional=True), "MATERIALIZED", optional=True),
        Bracketed(
            Ref("SelectableGrammar"),
            parse_mode=ParseMode.GREEDY,
        ),
    )



"""The capitalisation plugin bundle."""

from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.rules import BaseRule, ConfigInfo


@hookimpl
def get_configs_info() -> dict[str, ConfigInfo]:
    """Get additional rule config validations and descriptions."""
    return {
        "capitalisation_policy": {
            "validation": ["consistent", "upper", "lower", "capitalise"],
            "definition": "The capitalisation policy to enforce.",
        },
        "extended_capitalisation_policy": {
            "validation": [
                "consistent",
                "upper",
                "lower",
                "pascal",
                "capitalise",
                "snake",
                "camel",
            ],
            "definition": (
                "The capitalisation policy to enforce, extended with PascalCase, "
                "snake_case, and camelCase. "
                "This is separate from ``capitalisation_policy`` as it should not be "
                "applied to keywords."
                "Camel, Pascal, and Snake will never be inferred when the policy is "
                "set to consistent. This is because snake can cause destructive "
                "changes to the identifier, and unlinted code is too easily mistaken "
                "for camel and pascal. If, when set to consistent, no consistent "
                "case is found, it will default to upper."
            ),
        },
    }


@hookimpl
def get_rules() -> list[type[BaseRule]]:
    """Get plugin rules.

    NOTE: Rules are imported only on fetch to manage import times
    when rules aren't used.
    """
    from sqlfluff.rules.capitalisation.CP01 import Rule_CP01
    from sqlfluff.rules.capitalisation.CP02 import Rule_CP02
    from sqlfluff.rules.capitalisation.CP03 import Rule_CP03
    from sqlfluff.rules.capitalisation.CP04 import Rule_CP04
    from sqlfluff.rules.capitalisation.CP05 import Rule_CP05

    return [Rule_CP01, Rule_CP02, Rule_CP03, Rule_CP04, Rule_CP05]



"""A list of databricks reserved keywords.

https://docs.databricks.com/sql/language-manual/sql-ref-reserved-words.html
"""

RESERVED_KEYWORDS = [
    "ANTI",
    "CROSS",
    "EXCEPT",
    "FULL",
    "INNER",
    "INTERSECT",
    "JOIN",
    "LATERAL",
    "LEFT",
    "MINUS",
    "NATURAL",
    "ON",
    "RIGHT",
    "SEMI",
    "UNION",
    "USING",
]

UNRESERVED_KEYWORDS = [
    "AUTO",
    "CATALOG",
    "CDC",
    "CLONE",
    "COMPENSATION",
    "CRON",
    "DAY",
    "DEEP",
    "DAYS",
    "ENFORCED",
    "EVOLUTION",
    "FEATURE",
    "FILTER",
    "FLOW",
    "HOUR",
    "HOURS",
    "IDENTIFIER",
    "MANAGED",
    "MASK",
    "METRICS",
    "MOST",
    "NORELY",
    "OPTIMIZATION",
    "OPTIMIZE",
    "PREDICTIVE",
    "PRIVATE",
    "PROVIDER",
    "PYTHON",
    "RECIPIENT",
    "REFRESH",
    "RELY",
    "ROW",
    "SCHEDULE",
    "SHALLOW",
    "SOURCE",
    "SQL",
    "TAG",
    "TAGS",
    "TARGET",
    "TIMESERIES",
    "TRIGGER",
    "VOLUME",
    "VOLUMES",
    "WEEK",
    "WEEKS",
    "YAML",
    "ZORDER",
]



"""The Greenplum dialect.

Greenplum (http://www.greenplum.org/) is a Massively Parallel Postgres,
so we base this dialect on Postgres.
"""

from sqlfluff.core.dialects import load_raw_dialect
from sqlfluff.core.parser import (
    AnyNumberOf,
    AnySetOf,
    Anything,
    BaseSegment,
    Bracketed,
    Dedent,
    Delimited,
    Indent,
    OneOf,
    OptionallyBracketed,
    ParseMode,
    Ref,
    Sequence,
)
from sqlfluff.dialects import dialect_ansi as ansi
from sqlfluff.dialects import dialect_postgres as postgres
from sqlfluff.dialects.dialect_greenplum_keywords import greenplum_keywords
from sqlfluff.dialects.dialect_postgres_keywords import get_keywords

postgres_dialect = load_raw_dialect("postgres")

greenplum_dialect = postgres_dialect.copy_as(
    "greenplum",
    formatted_name="Greenplum",
    docstring="The dialect for `Greenplum <https://www.greenplum.org/>`_.",
)

greenplum_dialect.sets("reserved_keywords").update(
    get_keywords(greenplum_keywords, "reserved")
)

greenplum_dialect.sets("unreserved_keywords").update(
    get_keywords(greenplum_keywords, "non-reserved")
)


class StatementSegment(postgres.StatementSegment):
    """A generic segment, to any of its child subsegments."""

    match_grammar = postgres.StatementSegment.match_grammar.copy(
        insert=[
            Ref("FetchClauseSegment"),
            Ref("DeclareStatement"),
            Ref("CloseStatementSegment"),
            Ref("AnalizeSegment"),
            Ref("CreateExternalTableStatementSegment"),
            Ref("DropExternalTableStatementSegment"),
        ],
    )


class SelectClauseSegment(postgres.SelectClauseSegment):
    """Overrides Postgres to allow DISTRIBUTED as a terminator."""

    match_grammar = Sequence(
        "SELECT",
        Ref("SelectClauseModifierSegment", optional=True),
        Indent,
        Delimited(
            Ref("SelectClauseElementSegment"),
            # In Postgres you don't need an element so make it optional
            optional=True,
            allow_trailing=True,
        ),
        Dedent,
        terminators=[
            "INTO",
            "FROM",
            "WHERE",
            Sequence("ORDER", "BY"),
            "LIMIT",
            "OVERLAPS",
            Ref("SetOperatorSegment"),
            Sequence("WITH", Ref.keyword("NO", optional=True), "DATA"),
            Ref("WithCheckOptionSegment"),
            Ref("DistributedBySegment"),
        ],
        parse_mode=ParseMode.GREEDY_ONCE_STARTED,
    )


class DistributedBySegment(BaseSegment):
    """A DISTRIBUTED BY clause."""

    type = "distributed_by"

    match_grammar = Sequence(
        "DISTRIBUTED",
        OneOf(
            "RANDOMLY",
            "REPLICATED",
            Sequence("BY", Bracketed(Delimited(Ref("ColumnReferenceSegment")))),
        ),
    )


class CreateExternalTableStatementSegment(BaseSegment):
    """A `CREATE EXTERNAL TABLE` statement.

    As specified in
    https://docs.vmware.com/en/VMware-Greenplum/7/greenplum-database/ref_guide-sql_commands-CREATE_EXTERNAL_TABLE.html
    """

    type = "create_external_table_statement"

    match_grammar = Sequence(
        "CREATE",
        OneOf("READABLE", "WRITABLE", optional=True),
        "EXTERNAL",
        Ref.keyword("WEB", optional=True),
        "TABLE",
        Ref("TableReferenceSegment"),
        OneOf(
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ColumnReferenceSegment"),
                        Ref("DatatypeSegment"),
                    ),
                ),
            ),
            Sequence(
                "LIKE",
                Ref("TableReferenceSegment"),
            ),
        ),
        OneOf(
            Sequence(
                "LOCATION",
                Bracketed(
                    Delimited(
                        Ref("QuotedLiteralSegment"),
                    ),
                ),
            ),
            Sequence(
                "EXECUTE",
                Ref("QuotedLiteralSegment"),
                Sequence(
                    "ON",
                    OneOf(
                        "ALL",
                        Sequence("HOST", Ref("QuotedLiteralSegment", optional=True)),
                        Sequence(
                            "SEGMENT",
                            Ref("NumericLiteralSegment"),
                        ),
                        "COORDINATOR",
                        Ref("NumericLiteralSegment"),
                    ),
                    optional=True,
                ),
            ),
        ),
        "FORMAT",
        Ref("QuotedLiteralSegment"),
        Bracketed(
            Delimited(
                OneOf(
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("EqualsSegment"),
                        Ref("QuotedLiteralSegment"),
                    ),
                    Sequence(
                        Ref("ParameterNameSegment"),
                        OneOf(
                            Ref("QuotedLiteralSegment"),
                            Ref("QuotedIdentifierSegment"),
                            optional=True,
                        ),
                    ),
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref.keyword("AS", optional=True),
                        Ref("QuotedLiteralSegment"),
                    ),
                ),
            ),
            optional=True,
        ),
        Sequence(
            "OPTIONS",
            Bracketed(
                Delimited(
                    Sequence(
                        Ref("ParameterNameSegment"),
                        Ref("QuotedLiteralSegment"),
                    ),
                ),
            ),
            optional=True,
        ),
        AnyNumberOf(
            Sequence(
                "ENCODING",
                Ref("QuotedLiteralSegment"),
            ),
            Sequence(
                "LOG",
                "ERRORS",
                Ref.keyword("PERSISTENTLY", optional=True),
            ),
            Sequence(
                "SEGMENT",
                "REJECT",
                "LIMIT",
                Ref("NumericLiteralSegment"),
                OneOf("ROWS", "PERCENT", optional=True),
            ),
            Ref("DistributedBySegment"),
        ),
    )


class DropExternalTableStatementSegment(BaseSegment):
    """A `DROP EXTERNAL TABLE` statement.

    As specified in
    https://docs.vmware.com/en/VMware-Greenplum/7/greenplum-database/ref_guide-sql_commands-DROP_EXTERNAL_TABLE.html
    """

    type = "drop_external_table_statement"

    match_grammar = Sequence(
        "DROP",
        "EXTERNAL",
        Ref.keyword("WEB", optional=True),
        "TABLE",
        Ref("IfExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        Ref("DropBehaviorGrammar", optional=True),
    )


class CreateTableStatementSegment(postgres.CreateTableStatementSegment):
    """A `CREATE TABLE` statement.

    As specified in
    https://docs.vmware.com/en/VMware-Greenplum/7/greenplum-database/ref_guide-sql_commands-CREATE_TABLE.html
    This is overridden from Postgres to add the `DISTRIBUTED` clause.
    """

    match_grammar = Sequence(
        "CREATE",
        OneOf(
            Sequence(
                OneOf("GLOBAL", "LOCAL", optional=True),
                Ref("TemporaryGrammar", optional=True),
            ),
            "UNLOGGED",
            optional=True,
        ),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        OneOf(
            # Columns and comment syntax:
            Sequence(
                Bracketed(
                    Delimited(
                        OneOf(
                            Sequence(
                                Ref("ColumnReferenceSegment"),
                                Ref("DatatypeSegment"),
                                AnyNumberOf(
                                    # A single COLLATE segment can come before or after
                                    # constraint segments
                                    OneOf(
                                        Ref("ColumnConstraintSegment"),
                                        Sequence(
                                            "COLLATE",
                                            Ref("CollationReferenceSegment"),
                                        ),
                                    ),
                                ),
                            ),
                            Ref("TableConstraintSegment"),
                            Sequence(
                                "LIKE",
                                Ref("TableReferenceSegment"),
                                AnyNumberOf(Ref("LikeOptionSegment"), optional=True),
                            ),
                        ),
                    )
                ),
                Sequence(
                    "INHERITS",
                    Bracketed(Delimited(Ref("TableReferenceSegment"))),
                    optional=True,
                ),
            ),
            # Create OF syntax:
            Sequence(
                "OF",
                Ref("ParameterNameSegment"),
                Bracketed(
                    Delimited(
                        Sequence(
                            Ref("ColumnReferenceSegment"),
                            Sequence("WITH", "OPTIONS", optional=True),
                            AnyNumberOf(Ref("ColumnConstraintSegment")),
                        ),
                        Ref("TableConstraintSegment"),
                    ),
                    optional=True,
                ),
            ),
            # Create PARTITION OF syntax
            Sequence(
                "PARTITION",
                "OF",
                Ref("TableReferenceSegment"),
                Bracketed(
                    Delimited(
                        Sequence(
                            Ref("ColumnReferenceSegment"),
                            Sequence("WITH", "OPTIONS", optional=True),
                            AnyNumberOf(Ref("ColumnConstraintSegment")),
                        ),
                        Ref("TableConstraintSegment"),
                    ),
                    optional=True,
                ),
                OneOf(
                    Sequence("FOR", "VALUES", Ref("PartitionBoundSpecSegment")),
                    "DEFAULT",
                ),
            ),
        ),
        AnyNumberOf(
            Sequence(
                "PARTITION",
                "BY",
                OneOf("RANGE", "LIST"),
                Bracketed(
                    Ref("ColumnReferenceSegment"),
                ),
                AnyNumberOf(
                    Sequence(
                        "SUBPARTITION",
                        "BY",
                        OneOf("RANGE", "LIST"),
                        Bracketed(
                            Ref("ColumnReferenceSegment"),
                        ),
                        Sequence(
                            "SUBPARTITION",
                            "TEMPLATE",
                            Bracketed(
                                # TODO: Is this too permissive?
                                Anything(),
                            ),
                            optional=True,
                        ),
                    ),
                ),
                Bracketed(
                    # TODO: Is this too permissive?
                    Anything(),
                ),
            ),
            Sequence("USING", Ref("ParameterNameSegment")),
            Sequence(
                "WITH",
                Bracketed(
                    Delimited(
                        Sequence(
                            Ref("ParameterNameSegment"),
                            Sequence(
                                Ref("EqualsSegment"),
                                OneOf(
                                    Ref("LiteralGrammar"),
                                    Ref("NakedIdentifierSegment"),
                                    Ref("QuotedIdentifierSegment"),
                                ),
                                optional=True,
                            ),
                        ),
                    )
                ),
            ),
            Sequence(
                "ON",
                "COMMIT",
                OneOf(Sequence("PRESERVE", "ROWS"), Sequence("DELETE", "ROWS"), "DROP"),
            ),
            Sequence("TABLESPACE", Ref("TablespaceReferenceSegment")),
            Ref("DistributedBySegment"),
        ),
    )


class CreateTableAsStatementSegment(postgres.CreateTableAsStatementSegment):
    """A `CREATE TABLE AS` statement.

    As specified in
    https://docs.vmware.com/en/VMware-Greenplum/7/greenplum-database/ref_guide-sql_commands-CREATE_TABLE_AS.html
    This is overridden from Postgres to add the `DISTRIBUTED` clause.
    """

    match_grammar = Sequence(
        "CREATE",
        OneOf(
            Sequence(
                OneOf("GLOBAL", "LOCAL", optional=True),
                Ref("TemporaryGrammar"),
            ),
            "UNLOGGED",
            optional=True,
        ),
        "TABLE",
        Ref("IfNotExistsGrammar", optional=True),
        Ref("TableReferenceSegment"),
        AnyNumberOf(
            Bracketed(
                Delimited(Ref("ColumnReferenceSegment")),
                optional=True,
            ),
            Sequence("USING", Ref("ParameterNameSegment"), optional=True),
            OneOf(
                Sequence(
                    "WITH",
                    Bracketed(
                        Delimited(
                            Sequence(
                                Ref("ParameterNameSegment"),
                                Sequence(
                                    Ref("EqualsSegment"),
                                    OneOf(
                                        Ref("LiteralGrammar"),
                                        Ref("NakedIdentifierSegment"),
                                        Ref("QuotedIdentifierSegment"),
                                    ),
                                    optional=True,
                                ),
                            )
                        )
                    ),
                ),
                Sequence("WITHOUT", "OIDS"),
                optional=True,
            ),
            Sequence(
                "ON",
                "COMMIT",
                OneOf(Sequence("PRESERVE", "ROWS"), Sequence("DELETE", "ROWS"), "DROP"),
                optional=True,
            ),
            Sequence("TABLESPACE", Ref("TablespaceReferenceSegment"), optional=True),
        ),
        "AS",
        OneOf(
            OptionallyBracketed(Ref("SelectableGrammar")),
            OptionallyBracketed(Sequence("TABLE", Ref("TableReferenceSegment"))),
            Ref("ValuesClauseSegment"),
            OptionallyBracketed(Sequence("EXECUTE", Ref("FunctionSegment"))),
        ),
        Ref("WithDataClauseSegment", optional=True),
        Ref("DistributedBySegment", optional=True),
    )


class UnorderedSelectStatementSegment(postgres.UnorderedSelectStatementSegment):
    """Overrides Postgres Statement, adding DISTRIBUTED BY as a terminator."""

    match_grammar = postgres.UnorderedSelectStatementSegment.match_grammar.copy(
        terminators=[
            Ref("DistributedBySegment"),
        ],
    )


class SelectStatementSegment(postgres.SelectStatementSegment):
    """Overrides Postgres Statement, adding DISTRIBUTED BY as a terminator."""

    match_grammar = postgres.SelectStatementSegment.match_grammar.copy(
        terminators=[
            Ref("DistributedBySegment"),
        ],
    )


class AnalizeSegment(BaseSegment):
    """ANALYZE statement.

    https://docs.vmware.com/en/VMware-Greenplum/6/greenplum-database/ref_guide-sql_commands-ANALYZE.html
    """

    type = "analize_statement"

    match_grammar = Sequence(
        OneOf("ANALYZE", "ANALYSE"),
        Ref.keyword("VERBOSE", optional=True),
        Ref.keyword("ROOTPARTITION", optional=True),
        OneOf(
            Sequence(
                Ref("TableReferenceSegment"),
                Bracketed(
                    Delimited(
                        Ref("ColumnReferenceSegment"),
                        allow_trailing=True,
                    ),
                    optional=True,
                ),
            ),
            "ALL",
            optional=True,
        ),
    )


class FetchClauseSegment(ansi.FetchClauseSegment):
    """FETCH statement.

    https://docs.vmware.com/en/VMware-Greenplum/6/greenplum-database/ref_guide-sql_commands-FETCH.html
    """

    type = "fetch_clause"
    match_grammar = Sequence(
        "FETCH",
        Sequence(
            OneOf(
                "FIRST",
                "NEXT",
                Sequence("ABSOLUTE", Ref("NumericLiteralSegment")),
                Sequence("RELATIVE", Ref("NumericLiteralSegment")),
                Ref("NumericLiteralSegment"),
                "ALL",
                "FORWARD",
                Sequence("FORWARD", Ref("NumericLiteralSegment")),
                Sequence("FORWARD", "ALL"),
            ),
            OneOf("FROM", "IN"),
            optional=True,
        ),
        Ref("TableReferenceSegment"),
    )


class DeclareStatement(BaseSegment):
    """DECLARE statement.

    https://docs.vmware.com/en/VMware-Greenplum/6/greenplum-database/ref_guide-sql_commands-DECLARE.html
    """

    type = "declare_statement"

    match_grammar = Sequence(
        "DECLARE",
        Ref("TableReferenceSegment"),
        AnySetOf(
            Ref.keyword("BINARY", optional=True),
            Ref.keyword("INSENSITIVE", optional=True),
            Sequence(
                "NO",
                "SCROLL",
                optional=True,
            ),
            Sequence(
                "PARALLEL",
                "RETRIEVE",
                optional=True,
            ),
            optional=True,
        ),
        "CURSOR",
        Sequence(
            OneOf("WITH", "WITHOUT"),
            "HOLD",
            optional=True,
        ),
        "FOR",
        Ref("StatementSegment"),
        Sequence(
            "FOR",
            "READ",
            "ONLY",
            optional=True,
        ),
    )


class CloseStatementSegment(BaseSegment):
    """CLOSE statement.

    https://docs.vmware.com/en/VMware-Greenplum/7/greenplum-database/ref_guide-sql_commands-CLOSE.html
    """

    type = "close_statement"

    match_grammar = Sequence(
        "CLOSE",
        OneOf(Ref("TableReferenceSegment"), "ALL"),
    )


class CopyStatementSegment(postgres.CopyStatementSegment):
    """COPY statement.

    https://docs.vmware.com/en/VMware-Greenplum/6/greenplum-database/ref_guide-sql_commands-COPY.html
    """

    type = "copy_statement"

    _target_subset = OneOf(
        Ref("QuotedLiteralSegment"),
        Sequence("PROGRAM", Ref("QuotedLiteralSegment")),
    )

    _table_definition = Sequence(
        Ref("TableReferenceSegment"),
        Bracketed(
            Delimited(
                Ref("ColumnReferenceSegment"),
                allow_trailing=True,
            ),
            optional=True,
        ),
    )

    _option = Sequence(
        AnySetOf(
            Sequence("FORMAT", Ref("SingleIdentifierGrammar")),
            Sequence("ON", "SEGMENT"),
            "BINARY",
            Sequence("OIDS", Ref("BooleanLiteralGrammar", optional=True)),
            Sequence("FREEZE", Ref("BooleanLiteralGrammar", optional=True)),
            Sequence(
                "DELIMITER",
                Ref.keyword("AS", optional=True),
                Ref("QuotedLiteralSegment"),
            ),
            Sequence(
                "NULL",
                Ref.keyword("AS", optional=True),
                Ref("QuotedLiteralSegment"),
            ),
            Sequence("HEADER", Ref("BooleanLiteralGrammar", optional=True)),
            Sequence("QUOTE", Ref("QuotedLiteralSegment")),
            Sequence(
                "ESCAPE",
                Ref.keyword("AS", optional=True),
                Ref("QuotedLiteralSegment"),
            ),
            Sequence(
                "NEWLINE",
                Ref.keyword("AS", optional=True),
                Ref("QuotedLiteralSegment"),
            ),
            Sequence(
                "FORCE_QUOTE",
                OneOf(
                    Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    Ref("StarSegment"),
                ),
            ),
            Sequence(
                "FORCE_NOT_NULL",
                Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
            ),
            Sequence(
                "FORCE_NULL",
                Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
            ),
            Sequence("ENCODING", Ref("QuotedLiteralSegment")),
            Sequence("FILL", "MISSING", "FIELDS"),
            Sequence(
                "LOG",
                "ERRORS",
                Sequence(
                    "SEGMENT",
                    "REJECT",
                    "LIMIT",
                    Ref("NumericLiteralSegment"),
                    OneOf(
                        "ROWS",
                        "PERCENT",
                        optional=True,
                    ),
                    optional=True,
                ),
            ),
            Sequence(
                "CSV",
                Sequence(
                    "QUOTE",
                    Ref.keyword("AS", optional=True),
                    Ref("QuotedLiteralSegment"),
                    optional=True,
                ),
                OneOf(
                    Sequence(
                        "FORCE",
                        "NOT",
                        "NULL",
                        Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    ),
                    Sequence(
                        "FORCE",
                        "QUOTE",
                        Bracketed(Delimited(Ref("ColumnReferenceSegment"))),
                    ),
                    optional=True,
                ),
            ),
            Sequence("IGNORE", "EXTERNAL", "PARTITIONS"),
        ),
        optional=True,
    )

    _bracketed_option = Sequence(
        Bracketed(
            Delimited(
                _option,
            )
        )
    )

    match_grammar = Sequence(
        "COPY",
        OneOf(
            Sequence(
                _table_definition,
                "FROM",
                OneOf(
                    _target_subset,
                    Sequence("STDIN"),
                ),
                Ref.keyword("WITH", optional=True),
                OneOf(_option, _bracketed_option, optional=True),
                Sequence("ON", "SEGMENT", optional=True),
            ),
            Sequence(
                OneOf(
                    _table_definition,
                    Bracketed(Ref("UnorderedSelectStatementSegment")),
                ),
                "TO",
                OneOf(
                    _target_subset,
                    Sequence("STDOUT"),
                ),
                OneOf(
                    Sequence(
                        Ref.keyword("WITH", optional=True),
                        OneOf(_option, _bracketed_option, optional=True),
                    ),
                    Ref("StarSegment"),
                    optional=True,
                ),
                Sequence("ON", "SEGMENT", optional=True),
            ),
        ),
    )



"""Helper classes & methods for applying fixes to segments."""

import logging
from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

from sqlfluff.core.parser import BaseSegment, SourceFix
from sqlfluff.core.rules.fix import LintFix

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.dialects import Dialect


linter_logger = logging.getLogger("sqlfluff.linter")


@dataclass
class AnchorEditInfo:
    """For a given fix anchor, count of the fix edit types and fixes for it."""

    delete: int = field(default=0)
    replace: int = field(default=0)
    create_before: int = field(default=0)
    create_after: int = field(default=0)
    fixes: list["LintFix"] = field(default_factory=list)
    source_fixes: list[SourceFix] = field(default_factory=list)
    # First fix of edit_type "replace" in "fixes"
    _first_replace: Optional["LintFix"] = field(default=None)

    def add(self, fix: "LintFix") -> None:
        """Adds the fix and updates stats.

        We also allow potentially multiple source fixes on the same
        anchor by condensing them together here.
        """
        if fix in self.fixes:
            # Deduplicate fixes in case it's already in there.
            return

        if fix.is_just_source_edit():
            assert fix.edit
            # is_just_source_edit confirms there will be a list
            # so we can hint that to mypy.
            self.source_fixes += fix.edit[0].source_fixes
            # is there already a replace?
            if self._first_replace:
                assert self._first_replace.edit
                # is_just_source_edit confirms there will be a list
                # and that's the only way to get into _first_replace
                # if it's populated so we can hint that to mypy.
                linter_logger.info(
                    "Multiple edits detected, condensing %s onto %s",
                    fix,
                    self._first_replace,
                )
                self._first_replace.edit[0] = self._first_replace.edit[0].edit(
                    source_fixes=self.source_fixes
                )
                linter_logger.info("Condensed fix: %s", self._first_replace)
                # Return without otherwise adding in this fix.
                return

        self.fixes.append(fix)
        if fix.edit_type == "replace" and not self._first_replace:
            self._first_replace = fix
        setattr(self, fix.edit_type, getattr(self, fix.edit_type) + 1)

    @property
    def total(self) -> int:
        """Returns total count of fixes."""
        return len(self.fixes)

    @property
    def is_valid(self) -> bool:
        """Returns True if valid combination of fixes for anchor.

        Cases:
        * 0-1 fixes of any type: Valid
        * 2 fixes: Valid if and only if types are create_before and create_after
        """
        if self.total <= 1:
            # Definitely valid (i.e. no conflict) if 0 or 1. In practice, this
            # function probably won't be called if there are 0 fixes, but 0 is
            # valid; it simply means "no fixes to apply".
            return True
        if self.total == 2:
            # This is only OK for this special case. We allow this because
            # the intent is clear (i.e. no conflict): Insert something *before*
            # the segment and something else *after* the segment.
            return self.create_before == 1 and self.create_after == 1
        # Definitely bad if > 2.
        return False  # pragma: no cover


def compute_anchor_edit_info(fixes: list["LintFix"]) -> dict[int, AnchorEditInfo]:
    """Group and count fixes by anchor, return dictionary."""
    anchor_info = defaultdict(AnchorEditInfo)  # type: ignore
    for fix in fixes:
        # :TRICKY: Use segment uuid as the dictionary key since
        # different segments may compare as equal.
        anchor_id = fix.anchor.uuid
        anchor_info[anchor_id].add(fix)
    return dict(anchor_info)


def apply_fixes(
    segment: BaseSegment,
    dialect: "Dialect",
    rule_code: str,
    fixes: dict[int, AnchorEditInfo],
    max_parse_depth: int,
    max_parse_nodes: int = 0,
    fix_even_unparsable: bool = False,
) -> tuple["BaseSegment", list["BaseSegment"], list["BaseSegment"], bool]:
    """Apply a dictionary of fixes to this segment.

    Used in to apply fixes found in linting. If a segment remains unchanged
    then the original is returned, but if any changes are made to it, or any
    of it's child segments, then it returns a copy rather than mutating the
    original.

    Most fixes are usually applied when this method is called on their parent
    segment, this is because that's where we can insert or move segments relative
    to the anchor specified in the fix. This has the implication that if the
    method is called on a `RawSegment`, then no changes will be applied, because
    a `RawSegment` never has child segments.

    After fixing, it calls `validate_segment_with_reparse` on the segment to
    check that the segment still parses after any changes are made. The result
    of this is returned as a boolean in the last element of the return tuple.
    As the function recurses, if an inner element doesn't parse after fixing,
    then the outer segment will also be checked, and if found to parse successfully
    then the method returns `True` as valid. This is because sometimes the fixes
    change the structure enough that a wider reparse is necessary.

    Because of this validity checking, any unparsable sections are assumed
    unfixable (because we won't know if we're corrupting the SQL). The method
    will therefore return early without applying any fixes if the segment it's
    called on is unparsable (because we already know that validation check will
    fail already).

    If `fix_even_unparsable` is True, then we will still apply fixes to unparsable
    sections, but will do so *without validation*. That means that the final
    element of the return value will always return `True`, so that we don't interrupt
    the validity checking of any outer (parsable) sections.
    """
    if not fixes or segment.is_raw():
        return segment, [], [], True

    seg_buffer = []
    before = []
    after = []
    fixes_applied: list[LintFix] = []
    requires_validate = False

    for seg in segment.segments:
        # Look for uuid match.
        # This handles potential positioning ambiguity.
        anchor_info: Optional[AnchorEditInfo] = fixes.pop(seg.uuid, None)

        if anchor_info is None:
            # No fix matches here, just add the segment and move on.
            seg_buffer.append(seg)
            continue

        # Otherwise there is a fix match.
        seg_fixes = anchor_info.fixes
        if (
            len(seg_fixes) == 2 and seg_fixes[0].edit_type == "create_after"
        ):  # pragma: no cover
            # Must be create_before & create_after. Swap so the
            # "before" comes first.
            seg_fixes.reverse()

        for f in anchor_info.fixes:
            assert f.anchor.uuid == seg.uuid
            fixes_applied.append(f)
            linter_logger.debug(
                "Matched fix for %s against segment: %s -> %s",
                rule_code,
                f,
                seg,
            )

            # Deletes are easy.
            if f.edit_type == "delete":
                # We're just getting rid of this segment.
                requires_validate = True
                # NOTE: We don't add the segment in this case.
                continue

            # Otherwise it must be a replace or a create.
            assert f.edit_type in (
                "replace",
                "create_before",
                "create_after",
            ), f"Unexpected edit_type: {f.edit_type!r} in {f!r}"

            if f.edit_type == "create_after" and len(anchor_info.fixes) == 1:
                # in the case of a creation after that is not part
                # of a create_before/create_after pair, also add
                # this segment before the edit.
                seg_buffer.append(seg)

            # We're doing a replacement (it could be a single
            # segment or an iterable)
            assert f.edit, f"Edit {f.edit_type!r} requires `edit`."
            consumed_pos = False
            for s in f.edit:
                seg_buffer.append(s)
                # If one of them has the same raw representation
                # then the first that matches gets to take the
                # original position marker.
                if f.edit_type == "replace" and s.raw == seg.raw and not consumed_pos:
                    seg_buffer[-1].pos_marker = seg.pos_marker
                    consumed_pos = True

            # If we're just editing a segment AND keeping the type the
            # same then no need to validate. Otherwise we should
            # trigger a validation (e.g. for creations or
            # multi-replace).
            if not (
                f.edit_type == "replace"
                and len(f.edit) == 1
                and f.edit[0].class_types == seg.class_types
            ):
                requires_validate = True

            if f.edit_type == "create_before":
                # in the case of a creation before, also add this
                # segment on the end
                seg_buffer.append(seg)

    # Invalidate any caches
    segment.invalidate_caches()

    # If any fixes applied, do an intermediate reposition. When applying
    # fixes to children and then trying to reposition them, that recursion
    # may rely on the parent having already populated positions for any
    # of the fixes applied there first. This ensures those segments have
    # working positions to work with.
    if fixes_applied:
        assert segment.pos_marker
        seg_buffer = list(
            segment._position_segments(tuple(seg_buffer), parent_pos=segment.pos_marker)
        )

    # Then recurse (i.e. deal with the children) (Requeueing)
    seg_queue = seg_buffer
    seg_buffer = []
    for seg in seg_queue:
        s, pre, post, validated = apply_fixes(
            seg,
            dialect,
            rule_code,
            fixes,
            max_parse_depth=max_parse_depth,
            max_parse_nodes=max_parse_nodes,
        )
        # 'before' and 'after' will usually be empty. Only used when
        # lower-level fixes left 'seg' with non-code (usually
        # whitespace) segments as the first or last children. This is
        # generally not allowed (see the can_start_end_non_code field),
        # and these segments need to be "bubbled up" the tree.
        seg_buffer += pre + [s] + post
        # If we fail to validate a child segment, make sure to validate this
        # segment.
        if not validated:
            requires_validate = True

    # Most correct whitespace positioning will have already been handled
    # _however_, the exception is `replace` edits which match start or
    # end with whitespace. We also need to handle any leading or trailing
    # whitespace ejected from the any fixes applied to child segments.
    # Here we handle those by checking the start and end of the resulting
    # segment sequence for whitespace.
    # If we're left with any non-code at the end, trim them off and pass them
    # up to the parent segment for handling.
    if not segment.can_start_end_non_code:
        _idx = 0
        for _idx in range(0, len(seg_buffer)):
            if segment._is_code_or_meta(seg_buffer[_idx]):
                break
        before = seg_buffer[:_idx]
        seg_buffer = seg_buffer[_idx:]

        _idx = len(seg_buffer)
        for _idx in range(len(seg_buffer), 0, -1):
            if segment._is_code_or_meta(seg_buffer[_idx - 1]):
                break
        after = seg_buffer[_idx:]
        seg_buffer = seg_buffer[:_idx]

    # Reform into a new segment
    assert segment.pos_marker
    try:
        new_seg = segment.__class__(
            # Realign the segments within
            segments=segment._position_segments(
                tuple(seg_buffer), parent_pos=segment.pos_marker
            ),
            pos_marker=segment.pos_marker,
            # Pass through any additional kwargs
            **{k: getattr(segment, k) for k in segment.additional_kwargs},
        )
    except AssertionError as err:  # pragma: no cover
        # An AssertionError on creating a new segment is likely a whitespace
        # check fail. If possible add information about the fixes we tried to
        # apply, before re-raising.
        # NOTE: only available in python 3.11+.
        if hasattr(err, "add_note"):
            err.add_note(f" After applying fixes: {fixes_applied}.")
        raise err

    # Handle any necessary validation.
    if requires_validate:
        # Was it already unparsable?
        if "unparsable" in segment.descendant_type_set | segment.class_types:
            if fix_even_unparsable:
                # If we're fixing even unparsable sections, there's no point trying
                # to validate, it will always fail. We may still want to validate
                # other sections of the file though, so we should just declare *this*
                # part of the file to be all good.
                validated = True
            else:
                # It was already unparsable, but we're being asked to validate.
                # Don't any apply fixes from within this region and just return the
                # original segment.
                return segment, [], [], True
        # Otherwise only validate if there's a match_grammar. Otherwise we may get
        # strange results (for example with the BracketedSegment).
        elif hasattr(new_seg, "match_grammar"):
            validated = new_seg.validate_segment_with_reparse(
                dialect,
                max_parse_depth=max_parse_depth,
                max_parse_nodes=max_parse_nodes,
            )
    else:
        validated = not requires_validate
    # Return the new segment and any non-code that needs to bubble up
    # the tree.
    # NOTE: We pass on whether this segment has been validated. It's
    # very possible that our parsing here may fail depending on the
    # type of segment that has been replaced, but if not we rely on
    # a parent segment still being valid. If we get all the way up
    # to the root and it's still not valid - that's a problem.
    return new_seg, before, after, validated



"""Definitions for Grammar."""

from collections.abc import Sequence
from typing import Optional, Union

from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.grammar import Ref
from sqlfluff.core.parser.grammar.anyof import OneOf
from sqlfluff.core.parser.grammar.noncode import NonCodeMatcher
from sqlfluff.core.parser.match_algorithms import (
    longest_match,
    skip_start_index_forward_to_code,
)
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.matchable import Matchable
from sqlfluff.core.parser.segments import BaseSegment


class Delimited(OneOf):
    """Match an arbitrary number of elements separated by a delimiter.

    Note that if there are multiple elements passed in that they will be treated
    as different options of what can be delimited, rather than a sequence.
    """

    equality_kwargs: tuple[str, ...] = (
        "_elements",
        "optional",
        "allow_gaps",
        "delimiter",
        "allow_trailing",
        "terminator",
        "min_delimiters",
    )

    optional_delimiter: bool = False

    def __init__(
        self,
        *args: Union[Matchable, str],
        delimiter: Union[Matchable, str] = Ref("CommaSegment"),
        allow_trailing: bool = False,
        terminators: Sequence[Union[Matchable, str]] = (),
        reset_terminators: bool = False,
        min_delimiters: int = 0,
        bracket_pairs_set: str = "bracket_pairs",
        allow_gaps: bool = True,
        optional: bool = False,
    ) -> None:
        """Initialize the class object with the provided arguments.

        Args:
            *args (Union[Matchable, str]): Options for elements between delimiters. This
                is treated as a set of options rather than a sequence.
            delimiter (Union[Matchable, str], optional): Delimiter used for parsing.
                Defaults to Ref("CommaSegment").
            allow_trailing (bool, optional): Flag indicating whether trailing delimiters
                are allowed. Defaults to False.
            terminators (Sequence[Union[Matchable, str]], optional): Sequence of
                terminators used to match the end of a segment.
                Defaults to ().
            reset_terminators (bool, optional): Flag indicating whether terminators
                should be reset. Defaults to False.
            min_delimiters (int, optional): Minimum number of delimiters to
                match. Defaults to 0.
            bracket_pairs_set (str, optional): Name of the bracket pairs set. Defaults
                to "bracket_pairs".
            allow_gaps (bool, optional): Flag indicating whether gaps between segments
                are allowed. Defaults to True.
            optional (bool, optional): Flag indicating whether the segment is optional.
                Defaults to False.
        """
        if delimiter is None:  # pragma: no cover
            raise ValueError("Delimited grammars require a `delimiter`")
        self.bracket_pairs_set = bracket_pairs_set
        self.delimiter = self._resolve_ref(delimiter)
        self.allow_trailing = allow_trailing
        # Setting min delimiters means we have to match at least this number
        self.min_delimiters = min_delimiters
        super().__init__(
            *args,
            terminators=terminators,
            reset_terminators=reset_terminators,
            allow_gaps=allow_gaps,
            optional=optional,
        )

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match delimited sequences.

        To achieve this we flip flop between looking for content
        and looking for delimiters. Individual elements of this
        grammar are treated as _options_ not as a _sequence_.
        """
        delimiters = 0
        seeking_delimiter = False
        max_idx = len(segments)
        working_idx = idx
        working_match = MatchResult.empty_at(idx)
        delimiter_match: Optional[MatchResult] = None

        delimiter_matchers = [self.delimiter]
        # NOTE: If the configured delimiter is in `parse_context.terminators` then
        # treat is _only_ as a delimiter and not as a terminator. This happens
        # frequently during nested comma expressions.
        terminator_matchers = [
            *self.terminators,
            *(t for t in parse_context.terminators if t not in delimiter_matchers),
        ]

        # If gaps aren't allowed, a gap (or non-code segment), acts like a terminator.
        if not self.allow_gaps:
            terminator_matchers.append(NonCodeMatcher())

        while True:
            # If we're past the start and allowed gaps, work forward
            # through any gaps.
            if self.allow_gaps and working_idx > idx:
                working_idx = skip_start_index_forward_to_code(segments, working_idx)

            # Do we have anything left to match on?
            if working_idx >= max_idx:
                break

            # Check whether there is a terminator before checking for content
            with parse_context.deeper_match(name="Delimited-Term") as ctx:
                match, _ = longest_match(
                    segments=segments,
                    matchers=terminator_matchers,
                    idx=working_idx,
                    parse_context=ctx,
                )
            if match:
                break

            # Then match for content/delimiter as appropriate.
            _push_terminators = []
            if delimiter_matchers and not seeking_delimiter:
                _push_terminators = delimiter_matchers
            with parse_context.deeper_match(
                name="Delimited", push_terminators=_push_terminators
            ) as ctx:
                match, _ = longest_match(
                    segments=segments,
                    matchers=(
                        delimiter_matchers if seeking_delimiter else self._elements
                    ),
                    idx=working_idx,
                    parse_context=ctx,
                )

            if not match:
                if seeking_delimiter and self.optional_delimiter:
                    # Failed to match a delimiter, but it's optional, so loop again.
                    seeking_delimiter = False
                    continue
                else:
                    # Failed to match next element, stop here.
                    break

            # Otherwise we _did_ match. Handle it.
            if seeking_delimiter:
                # It's a delimiter
                delimiter_match = match
            else:
                # It's content. Add both the last delimiter and the content to the
                # working match.
                if delimiter_match:
                    # NOTE: This should happen on every loop _except_ the first.
                    delimiters += 1
                    working_match = working_match.append(delimiter_match)
                working_match = working_match.append(match)

            # Prep for going back around the loop...
            working_idx = match.matched_slice.stop
            seeking_delimiter = not seeking_delimiter
            parse_context.update_progress(working_idx)

        if self.allow_trailing and delimiter_match and not seeking_delimiter:
            delimiters += 1
            working_match = working_match.append(delimiter_match)

        if delimiters < self.min_delimiters:
            return MatchResult.empty_at(idx)

        return working_match


class OptionallyDelimited(Delimited):
    """Match a number of elements optionally separated by a delimiter.

    Note that if there are multiple elements passed in that they will be treated
    as different options of what can be delimited, rather than a sequence.
    """

    optional_delimiter: bool = True



"""Implementation of Rule ST05."""

from collections.abc import Iterator
from functools import partial
from typing import NamedTuple, Optional, TypeVar, cast

from sqlfluff.core.dialects.base import Dialect
from sqlfluff.core.dialects.common import (
    AliasInfo,
    ObjectReferenceLevel,
    extract_possible_references,
)
from sqlfluff.core.parser import (
    BaseSegment,
    CodeSegment,
    KeywordSegment,
    NewlineSegment,
    SymbolSegment,
    WhitespaceSegment,
)
from sqlfluff.core.rules import (
    BaseRule,
    EvalResultType,
    LintFix,
    LintResult,
    RuleContext,
)
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.dialects.dialect_ansi import (
    CTEDefinitionSegment,
    TableExpressionSegment,
    TableReferenceSegment,
    WithCompoundStatementSegment,
)
from sqlfluff.utils.analysis.query import Query, Selectable
from sqlfluff.utils.analysis.select import get_select_statement_info
from sqlfluff.utils.functional import FunctionalContext, Segments
from sqlfluff.utils.functional.segment_predicates import (
    is_keyword,
    is_type,
    is_whitespace,
)

_SELECT_TYPES = [
    "with_compound_statement",
    "set_expression",
    "select_statement",
]


class _NestedSubQuerySummary(NamedTuple):
    query: Query
    selectable: Selectable
    table_alias: AliasInfo
    select_source_names: set[str]


class Rule_ST05(BaseRule):
    """Join/From clauses should not contain subqueries. Use CTEs instead.

    By default this rule is configured to allow subqueries within ``FROM``
    clauses but not within ``JOIN`` clauses. If you prefer a stricter lint
    then this is configurable.

    .. note::
       Some dialects don't allow CTEs, and for those dialects
       this rule makes no sense and should be disabled.

    **Anti-pattern**

    .. code-block:: sql

        select
            a.x, a.y, b.z
        from a
        join (
            select x, z from b
        ) using(x)


    **Best practice**

    .. code-block:: sql

        with c as (
            select x, z from b
        )
        select
            a.x, a.y, c.z
        from a
        join c using(x)

    """

    name = "structure.subquery"
    aliases = ("L042",)
    groups = ("all", "structure")
    config_keywords = ["forbid_subquery_in"]
    crawl_behaviour = SegmentSeekerCrawler(set(_SELECT_TYPES))

    _config_mapping = {
        "join": ["join_clause"],
        "from": ["from_expression_element"],
        "both": ["join_clause", "from_expression_element"],
    }
    is_fix_compatible = True

    # These are dialects that support WITH ... INSERT ... SELECT instead of
    # INSERT ... WITH ... SELECT
    # NOTE: this may be incomplete
    # NOTE: postgres supports both ways, so I've not included it here.
    _with_before_insert = {"tsql"}

    def _eval(self, context: RuleContext) -> EvalResultType:
        """Join/From clauses should not contain subqueries. Use CTEs instead."""
        self.forbid_subquery_in: str
        functional_context = FunctionalContext(context)
        segment = functional_context.segment
        parent_stack = functional_context.parent_stack
        is_select = segment.all(is_type(*_SELECT_TYPES))
        is_select_child = parent_stack.any(is_type(*_SELECT_TYPES))
        insert_parent = parent_stack.last(is_type("insert_statement"))
        if not is_select or is_select_child:
            # Nothing to do.
            return None

        query: Query = Query.from_segment(context.segment, context.dialect)

        is_with = segment.all(is_type("with_compound_statement"))
        # TODO: consider if we can fix recursive CTEs
        is_recursive = is_with and len(segment.children(is_keyword("recursive"))) > 0
        case_preference = _get_case_preference(segment)
        output_select = segment
        if is_with:
            output_select = segment.children(
                is_type(
                    "set_expression",
                    "select_statement",
                    "insert_statement",
                )
            )
        elif insert_parent and context.dialect.name in self._with_before_insert:
            # Here we select the parent `insert_statement` because it should be where
            # we place the new CTE.
            output_select = insert_parent
            segment = insert_parent

        # generate an instance which will track and shape our output CTE
        ctes = _CTEBuilder()
        # Init the output/final select & populate existing CTEs.
        # For WITH compound statements, collect all CTEs in document order,
        # including expression CTEs (e.g. ClickHouse ``expr AS alias``) that
        # are not captured by query.ctes (which only handles selectable CTEs).
        # For non-WITH statements (select_statement, set_expression), query.ctes
        # is always empty because Query.from_segment only populates ctes for
        # with_compound_statement segments, so there is nothing to pre-populate.
        if is_with:
            for cte_seg in context.segment.recursive_crawl(
                "common_table_expression",
                recurse_into=False,
                no_recursive_seg_type="with_compound_statement",
            ):
                ctes.insert_cte(cast(CTEDefinitionSegment, cte_seg))

        # Issue 3617: In T-SQL (and possibly other dialects) the automated fix
        # leaves parentheses in a location that causes a syntax error. This is an
        # unusual corner case. For simplicity, we still generate the lint warning
        # but don't try to generate a fix. Someone could look at this later (a
        # correct fix would involve removing the parentheses.)
        bracketed_ctas = [seg.type for seg in parent_stack[-2:]] == [
            "create_table_statement",
            "bracketed",
        ]

        # If there are offending elements calculate fixes
        clone_map = SegmentCloneMap(segment[0])
        results = self._lint_query(
            dialect=context.dialect,
            query=query,
            ctes=ctes,
            case_preference=case_preference,
            clone_map=clone_map,
        )

        results_list: list[tuple[LintResult, BaseSegment, str, BaseSegment, bool]] = []
        for result in results:
            (
                lint_result,
                from_expression,
                alias_name,
                subquery_parent,
                is_fixable,
            ) = result
            assert any(
                from_expression is seg for seg in subquery_parent.recursive_crawl_all()
            )
            results_list.append(result)
            if not is_fixable:
                continue
            this_seg_clone = clone_map[from_expression]
            new_table_ref = _create_table_ref(alias_name, context.dialect)
            # Add positions to the new table reference, other rules may need a position
            # but the clone is not a typical "fix".
            assert this_seg_clone.pos_marker
            this_seg_clone.segments = this_seg_clone._position_segments(
                (new_table_ref,), this_seg_clone.pos_marker
            )
            ctes.replace_with_clone(subquery_parent, clone_map)

        for (
            lint_result,
            from_expression,
            alias_name,
            subquery_parent,
            is_fixable,
        ) in results_list:
            if bracketed_ctas or is_recursive or not is_fixable:
                continue
            # Compute fix.
            output_select_clone = clone_map[output_select[0]]
            fixes = ctes.ensure_space_after_from(
                output_select[0], output_select_clone, subquery_parent
            )
            new_select = ctes.compose_select(
                output_select_clone, case_preference=case_preference
            )
            lint_result.fixes = [
                LintFix.replace(
                    segment[0],
                    edit_segments=[new_select],
                )
            ]
            lint_result.fixes += fixes
        return [lint_result[0] for lint_result in results_list]

    def _nested_subqueries(
        self, query: Query, dialect: Dialect
    ) -> Iterator[_NestedSubQuerySummary]:
        parent_types = self._config_mapping[self.forbid_subquery_in]
        for i, q in enumerate([query] + list(query.ctes.values())):
            for selectable in q.selectables:
                if not selectable.select_info:
                    continue  # pragma: no cover
                select_source_names = set()
                for a in selectable.select_info.table_aliases:
                    # For each table in FROM, return table name and any alias.
                    if a.ref_str:
                        select_source_names.add(a.ref_str)
                    if a.object_reference:
                        select_source_names.add(a.object_reference.raw)
                for table_alias in selectable.select_info.table_aliases:
                    try:
                        query = Query.from_root(
                            table_alias.from_expression_element, dialect
                        )
                    except AssertionError:
                        # Couldn't find a selectable, carry on.
                        continue

                    path_to = selectable.selectable.path_to(
                        table_alias.from_expression_element
                    )
                    if not (
                        # The from_expression_element
                        table_alias.from_expression_element.is_type(*parent_types)
                        # Or any of it's parents up to the selectable
                        or any(ps.segment.is_type(*parent_types) for ps in path_to)
                    ):
                        continue
                    # A set expression has one selectable per branch.
                    if any(
                        _is_correlated_subquery(
                            Segments(selectable_.selectable),
                            select_source_names,
                            dialect,
                        )
                        for selectable_ in query.selectables
                    ):
                        continue
                    yield _NestedSubQuerySummary(
                        q, selectable, table_alias, select_source_names
                    )
                    # Recursively find nested queries in CTEs
                    if i > 0:
                        yield from self._nested_subqueries(query, dialect)

    def _lint_query(
        self,
        dialect: Dialect,
        query: Query,
        ctes: "_CTEBuilder",
        case_preference: str,
        clone_map,
    ) -> Iterator[tuple[LintResult, BaseSegment, str, BaseSegment, bool]]:
        """Given the root query, compute lint warnings."""
        nsq: _NestedSubQuerySummary
        for nsq in self._nested_subqueries(query, dialect):
            alias_name, _ = ctes.create_cte_alias(nsq.table_alias)
            # 'anchor' is the TableExpressionSegment we fix/replace w/CTE name.
            anchor = nsq.table_alias.from_expression_element.segments[0]
            # If we have duplicate CTE names just don't fix anything
            # Return the lint warnings anyway
            is_fixable = alias_name not in ctes.list_used_names()

            # if the subquery is table_expression, get the bracketed child instead.
            if anchor.is_type("table_expression"):
                bracket_anchor = anchor.get_child("bracketed")
                # if the table_expression isn't bracketed, assume it isn't a subquery.
                if not bracket_anchor:
                    continue
            else:
                bracket_anchor = anchor

            # we can't create a CTE from a nested subquery here, ignore it.
            if not bracket_anchor.is_type("bracketed") or bracket_anchor.get_child(
                "table_expression"
            ):
                is_fixable = False
            if is_fixable:
                new_cte = _create_cte_seg(  # 'prep_1 as (select ...)'
                    alias_name=alias_name,
                    subquery=clone_map[bracket_anchor],
                    case_preference=case_preference,
                    dialect=dialect,
                )
                ctes.insert_cte(new_cte)

            # Grab the first keyword or symbol in the subquery to
            # use as the anchor. This makes the lint warning less
            # likely to be filtered out if a bit of the subquery
            # happens to be templated.
            anchor = next(anchor.recursive_crawl("keyword", "symbol"))
            res = LintResult(
                anchor=anchor,
                description=f"{nsq.selectable.selectable.type} clauses "
                "should not contain subqueries. Use CTEs instead",
                fixes=[],
            )
            yield (
                res,
                # FromExpressionElementSegment, parent of original "anchor" segment
                nsq.table_alias.from_expression_element,
                alias_name,  # Name of CTE we're creating from the nested query
                # Query with the subquery: 'select a from (select x from b)'
                nsq.selectable.selectable,
                is_fixable,
            )


def _get_first_select_statement_descendant(
    segment: BaseSegment,
) -> Optional[BaseSegment]:
    """Find first SELECT statement segment (if any) in descendants of 'segment'."""
    for select_statement in segment.recursive_crawl(
        "select_statement", recurse_into=False
    ):
        # We only want the first one.
        return select_statement
    return None  # pragma: no cover


def _is_correlated_subquery(
    nested_select: Segments, select_source_names: set[str], dialect: Dialect
) -> bool:
    """Given nested select and the sources of its parent, determine if correlated.

    https://en.wikipedia.org/wiki/Correlated_subquery
    """
    select_statement = _get_first_select_statement_descendant(nested_select[0])
    if not select_statement:
        return False  # pragma: no cover
    nested_select_info = get_select_statement_info(select_statement, dialect)
    if nested_select_info:
        for r in nested_select_info.reference_buffer:
            for tr in extract_possible_references(
                r, level=ObjectReferenceLevel.TABLE, dialect_name=dialect.name
            ):
                # Check for correlated subquery, as indicated by use of a
                # parent reference.
                if tr.part in select_source_names:
                    return True
    return False


class _CTEBuilder:
    """Gather CTE parts, maintain order and track naming/aliasing."""

    def __init__(self) -> None:
        self.ctes: list[CTEDefinitionSegment] = []
        self.name_idx = 0

    def list_used_names(self) -> list[str]:
        """Check CTEs and return used aliases."""
        used_names: list[str] = []
        for cte in self.ctes:
            id_seg = cte.get_identifier()
            cte_name = id_seg.raw
            if id_seg.is_type("quoted_identifier"):
                cte_name = cte_name[1:-1]

            used_names.append(cte_name)
        return used_names

    def insert_cte(self, cte: CTEDefinitionSegment) -> None:
        """Add a new CTE to the list as late as possible but before all its parents."""
        # This should still have the position markers of its true position
        inbound_subquery = (
            Segments(cte).children().last(lambda seg: bool(seg.pos_marker))
        )
        insert_position = next(
            (
                i
                for i, el in enumerate(self.ctes)
                if _is_child(Segments(el).children().last(), inbound_subquery)
            ),
            len(self.ctes),
        )

        self.ctes.insert(insert_position, cte)

    def create_cte_alias(self, alias: Optional[AliasInfo]) -> tuple[str, bool]:
        """Find or create the name for the next CTE."""
        if alias and alias.aliased and alias.ref_str:
            # If we know the name use it
            return alias.ref_str, False

        self.name_idx = self.name_idx + 1
        name = f"prep_{self.name_idx}"
        if name in self.list_used_names():
            # corner case where prep_x exists in origin query
            return self.create_cte_alias(None)
        return name, True

    def get_cte_segments(self) -> list[BaseSegment]:
        """Return a valid list of CTES with required padding segments."""
        cte_segments: list[BaseSegment] = []
        for cte in self.ctes:
            cte_segments += [
                cte,
                SymbolSegment(",", type="comma"),
                NewlineSegment(),
            ]
        return cte_segments[:-2]

    def compose_select(
        self, output_select_clone: BaseSegment, case_preference: str
    ) -> BaseSegment:
        """Compose our final new CTE."""
        # Compose the CTE.
        new_select = WithCompoundStatementSegment(
            segments=tuple(
                [
                    _segmentify("WITH", case_preference),
                    WhitespaceSegment(),
                    *self.get_cte_segments(),
                    NewlineSegment(),
                    output_select_clone,
                ]
            )
        )
        return new_select

    def ensure_space_after_from(
        self,
        output_select: BaseSegment,
        output_select_clone: BaseSegment,
        subquery_parent: BaseSegment,
    ) -> list[LintFix]:
        """Ensure there's whitespace between "FROM" and the CTE table name."""
        fixes = []
        if subquery_parent is output_select:
            (
                missing_space_after_from,
                from_clause,
                from_clause_children,
                from_segment,
            ) = self._missing_space_after_from(output_select_clone)
            if missing_space_after_from:
                # Case 1: from_clause is a child of cloned "output_select_clone"
                # that will be inserted by a fix. We can directly manipulate the
                # "segments" list. to insert whitespace between "FROM" and the
                # CTE table name.
                idx_from = from_clause_children.index(from_segment[0])
                from_clause.segments = list(
                    from_clause_children[: idx_from + 1]
                    + (WhitespaceSegment(),)
                    + from_clause_children[idx_from + 1 :]
                )
        else:
            (
                missing_space_after_from,
                from_clause,
                from_clause_children,
                from_segment,
            ) = self._missing_space_after_from(subquery_parent)
            if missing_space_after_from:
                # Case 2. from_segment is in the current parse tree, so we can't
                # modify it directly. Create a LintFix to do it.
                fixes.append(
                    LintFix.create_after(from_segment[0], [WhitespaceSegment()])
                )
        return fixes

    @staticmethod
    def _missing_space_after_from(segment: BaseSegment):
        missing_space_after_from = False
        from_clause_children = None
        from_segment = None
        from_clause = segment.get_child("from_clause")
        if from_clause is not None:
            from_clause_children = Segments(*from_clause.segments)
            from_segment = from_clause_children.first(is_keyword("from"))
            if from_segment and not from_clause_children.select(
                start_seg=from_segment[0], loop_while=is_whitespace()
            ):
                missing_space_after_from = True
        return missing_space_after_from, from_clause, from_clause_children, from_segment

    def replace_with_clone(self, segment, clone_map) -> None:
        for idx, cte in enumerate(self.ctes):
            if any(segment is seg for seg in cte.recursive_crawl_all()):
                self.ctes[idx] = clone_map[self.ctes[idx]]
                return None


def _is_child(maybe_parent: Segments, maybe_child: Segments) -> bool:
    """Is the child actually between the start and end markers of the parent."""
    assert len(maybe_child) == 1, (
        "Cannot assess child relationship of multiple segments"
    )
    assert len(maybe_parent) == 1, (
        "Cannot assess child relationship of multiple parents"
    )
    child_markers = maybe_child[0].pos_marker
    parent_pos = maybe_parent[0].pos_marker
    assert parent_pos and child_markers
    if child_markers.start_point_marker() < parent_pos.start_point_marker():
        return False  # pragma: no cover

    if child_markers.end_point_marker() > parent_pos.end_point_marker():
        return False

    return True


S = TypeVar("S", bound=type[BaseSegment])


def _get_seg(class_def: S, dialect: Dialect) -> S:
    return cast(S, dialect.get_segment(class_def.__name__))


def _create_cte_seg(
    alias_name: str, subquery: BaseSegment, case_preference: str, dialect: Dialect
) -> CTEDefinitionSegment:
    CTESegment = _get_seg(CTEDefinitionSegment, dialect)
    IdentifierSegment = cast(
        type[CodeSegment], dialect.get_segment("IdentifierSegment")
    )
    element: CTEDefinitionSegment = CTESegment(
        segments=(
            IdentifierSegment(
                raw=alias_name,
                type="naked_identifier",
            ),
            WhitespaceSegment(),
            _segmentify("AS", casing=case_preference),
            WhitespaceSegment(),
            # Return the bracketed segment instead of the table expression
            subquery,
        )
    )
    return element


def _create_table_ref(table_name: str, dialect: Dialect) -> TableExpressionSegment:
    Seg = partial(_get_seg, dialect=dialect)
    TableExpressionSeg = Seg(TableExpressionSegment)
    TableReferenceSeg = Seg(TableReferenceSegment)
    IdentifierSegment = cast(
        type[CodeSegment], dialect.get_segment("IdentifierSegment")
    )
    return TableExpressionSeg(
        segments=(
            TableReferenceSeg(
                segments=(
                    IdentifierSegment(
                        raw=table_name,
                        type="naked_identifier",
                    ),
                ),
            ),
        ),
    )


def _get_case_preference(root_select: Segments):
    # First get the segment itself so we have access to the generator
    root_segment = root_select.get()
    assert root_segment, "Root SELECT not found."
    # Get the first item of the recursive crawl.
    first_keyword = next(
        root_segment.recursive_crawl(
            "keyword",
            recurse_into=False,
        ),
        None,
    )
    assert first_keyword, "Keyword not found."
    # Get case preference based on the case of that keyword.
    if first_keyword.raw.islower():
        return "LOWER"
    return "UPPER"


def _segmentify(input_el: str, casing: str) -> BaseSegment:
    """Apply casing and convert strings to Keywords."""
    input_el = input_el.lower()
    if casing == "UPPER":
        input_el = input_el.upper()

    return KeywordSegment(raw=input_el)


class SegmentCloneMap:
    """Clones a segment tree, maps from original segments to their clones.

    The clone (a full ``segment.copy()``) is built lazily on first lookup: the
    caller builds the map for every SELECT statement but only *uses* it when
    there are subqueries to rewrite, so deferring avoids cloning the whole
    statement tree on the common no-subquery path (a notable cost over the Rust
    arena façade, where copy() materialises real segments).
    """

    def __init__(self, segment: BaseSegment):
        self._segment = segment
        self.segment_map: Optional[dict[int, BaseSegment]] = None

    def _build(self) -> dict[int, BaseSegment]:
        segment = self._segment
        segment_copy = segment.copy()
        segment_map: dict[int, BaseSegment] = {}
        for old_segment, new_segment in zip(
            segment.recursive_crawl_all(),
            segment_copy.recursive_crawl_all(),
        ):
            new_segment.pos_marker = old_segment.pos_marker
            segment_map[id(old_segment)] = new_segment
        self.segment_map = segment_map
        return segment_map

    def __getitem__(self, old_segment: BaseSegment) -> BaseSegment:
        segment_map = self.segment_map
        if segment_map is None:
            segment_map = self._build()
        return segment_map[id(old_segment)]



"""The ambiguous plugin bundle.

NOTE: Yes the title of this bundle is ...ambiguous. 😁
"""

from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.rules import BaseRule, ConfigInfo


@hookimpl
def get_configs_info() -> dict[str, ConfigInfo]:
    """Get additional rule config validations and descriptions."""
    return {
        "fully_qualify_join_types": {
            "validation": ["inner", "outer", "both"],
            "definition": ("Which types of JOIN clauses should be fully qualified?"),
        },
        "group_by_and_order_by_style": {
            "validation": ["consistent", "implicit", "explicit"],
            "definition": (
                "The expectation for using explicit column name references "
                "or implicit positional references."
            ),
        },
    }


@hookimpl
def get_rules() -> list[type[BaseRule]]:
    """Get plugin rules.

    NOTE: Rules are imported only on fetch to manage import times
    when rules aren't used.
    """
    from sqlfluff.rules.ambiguous.AM01 import Rule_AM01
    from sqlfluff.rules.ambiguous.AM02 import Rule_AM02
    from sqlfluff.rules.ambiguous.AM03 import Rule_AM03
    from sqlfluff.rules.ambiguous.AM04 import Rule_AM04
    from sqlfluff.rules.ambiguous.AM05 import Rule_AM05
    from sqlfluff.rules.ambiguous.AM06 import Rule_AM06
    from sqlfluff.rules.ambiguous.AM07 import Rule_AM07
    from sqlfluff.rules.ambiguous.AM08 import Rule_AM08
    from sqlfluff.rules.ambiguous.AM09 import Rule_AM09

    return [
        Rule_AM01,
        Rule_AM02,
        Rule_AM03,
        Rule_AM04,
        Rule_AM05,
        Rule_AM06,
        Rule_AM07,
        Rule_AM08,
        Rule_AM09,
    ]



"""String Helpers for the parser module."""

from collections.abc import Iterator
from typing import Union


def curtail_string(s: str, length: int = 20) -> str:
    """Trim a string nicely to length."""
    if len(s) > length:
        return s[:length] + "..."  # pragma: no cover
    else:
        return s


def findall(substr: str, in_str: str) -> Iterator[int]:
    """Yields all the positions sbstr within in_str.

    https://stackoverflow.com/questions/4664850/how-to-find-all-occurrences-of-a-substring
    """
    # Return nothing if one of the inputs is trivial
    if not substr or not in_str:
        return
    idx = in_str.find(substr)
    while idx != -1:
        yield idx
        idx = in_str.find(substr, idx + 1)


def split_colon_separated_string(in_str: str) -> tuple[tuple[str, ...], str]:
    r"""Converts a colon separated string.

    The final value in the string is handled separately the other others.
    >>> split_colon_separated_string("a:b")
    (('a',), 'b')
    >>> split_colon_separated_string("a:b:c")
    (('a', 'b'), 'c')
    >>> split_colon_separated_string("a:b:c:d")
    (('a', 'b', 'c'), 'd')
    >>> split_colon_separated_string("a")
    ((), 'a')

    NOTE: This also includes some heuristics for legit values containing colon.
    >>> split_colon_separated_string("foo:bar:C:\\Users")
    (('foo', 'bar'), 'C:\\Users')
    >>> split_colon_separated_string('foo:bar:{"k":"v"}')
    (('foo', 'bar'), '{"k":"v"}')
    >>> split_colon_separated_string('foo:bar:[{"k":"v"}]')
    (('foo', 'bar'), '[{"k":"v"}]')

    A two-character value ending in a colon is not a Windows path.
    >>> split_colon_separated_string("foo:a:")
    (('foo', 'a'), '')
    """
    config_path: list[str] = []
    leftover = in_str
    while ":" in leftover:
        element, _, value = leftover.partition(":")
        element = element.strip()
        value = value.strip()
        config_path.append(element)
        leftover = value
        if not should_split_on_colon(value):
            break

    # last part - actual value
    config_path.append(leftover)
    return tuple(config_path[:-1]), config_path[-1]


def should_split_on_colon(value: str) -> bool:
    """Heuristic for legit values containing comma."""
    if len(value) >= 3 and value[1] == ":" and value[2] == "\\":
        # Likely a Windows path
        return False
    if len(value) >= 2 and (value[0] == "[" and value[-1] == "]"):
        # Likely a JSON array
        return False
    if len(value) >= 2 and (value[0] == "{" and value[-1] == "}"):
        # Likely a JSON object
        return False

    return True


def split_comma_separated_string(raw: Union[str, list[str]]) -> list[str]:
    """Converts comma separated string to List, stripping whitespace."""
    if isinstance(raw, str):
        return [s.strip() for s in raw.split(",") if s.strip()]
    assert isinstance(raw, list)
    return raw


def get_trailing_whitespace_from_string(in_str: str) -> str:
    r"""Returns the trailing whitespace from a string.

    Designed to work with source strings of placeholders.

    >>> get_trailing_whitespace_from_string("")
    ''
    >>> get_trailing_whitespace_from_string("foo")
    ''
    >>> get_trailing_whitespace_from_string("   ")
    '   '
    >>> get_trailing_whitespace_from_string("  foo ")
    ' '
    >>> get_trailing_whitespace_from_string("foo\n")
    '\n'
    >>> get_trailing_whitespace_from_string("bar  \t  \n  \r ")
    '  \t  \n  \r '
    """
    whitespace_chars = " \t\r\n"
    if not in_str or in_str[-1] not in whitespace_chars:
        return ""  # No whitespace
    for i in range(1, len(in_str)):
        if in_str[-(i + 1)] not in whitespace_chars:
            # NOTE: The partial whitespace case is included as
            # future-proofing. In testing it appears it is never
            # required, and so only covered in the doctests above.
            # doctest coverage isn't included in the overall coverage
            # check and so the line below is excluded.
            return in_str[-i:]  # pragma: no cover
    else:
        return in_str  # All whitespace
